# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test package manager base."""
import copy
import json
import logging
import os
import re
import shutil
import tempfile
from collections import OrderedDict
from pathlib import Path
from unittest import mock

import pytest

from aea.cli.packages import package_type_selector_prompt
from aea.configurations.constants import PACKAGES
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.helpers.ipfs.base import IPFSHashOnly
from aea.package_manager.base import (
    DepedencyMismatchErrors,
    PackageFileNotValid,
    PackageHashDoesNotMatch,
    PackageNotValid,
    PackagesSourceNotValid,
)
from aea.package_manager.v1 import PackageManagerV1
from aea.protocols.generator.common import INIT_FILE_NAME
from aea.test_tools.test_cases import BaseAEATestCase

from tests.conftest import ROOT_DIR
from tests.test_package_manager.test_base import (
    DUMMY_PACKAGE_HASH,
    DUMMY_PACKAGE_ID,
    EXAMPLE_PACKAGE_HASH,
    EXAMPLE_PACKAGE_ID,
    PACKAGE_JSON_FILE,
)


TEST_SKILL_ID = PackageId(
    package_type=PackageType.SKILL,
    public_id=PublicId.from_str(
        "valory/abstract_round_abci:0.1.0:bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
    ),
)


class TestPackageManagerV1(BaseAEATestCase):
    """Test aea package manager."""

    use_packages_dir: bool = True
    packages_json_file: Path

    @classmethod
    def setup_class(cls) -> None:
        """Setup class."""

        super().setup_class()
        cls.packages_json_file = cls.t / PACKAGES / "packages.json"


class TestPackageManagerV1Initialization(TestPackageManagerV1):
    """Test object initialization."""

    def test_initialization(
        self,
    ) -> None:
        """Test object initialization."""

        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        packages = json.loads(self.packages_json_file.read_text())

        assert len(pm.dev_packages) == len(packages["dev"])
        assert len(pm.third_party_packages) == len(packages["third_party"])

        assert pm.path == self.packages_dir_path
        assert pm.get_package_hash(package_id=EXAMPLE_PACKAGE_ID) is not None, packages
        assert pm.get_package_hash(package_id=DUMMY_PACKAGE_ID) is None

        # fails on bad file
        with mock.patch.object(
            PackageManagerV1, "_load_packages", return_value={"some": []}
        ):
            with pytest.raises(PackageFileNotValid):
                PackageManagerV1.from_dir(self.packages_dir_path)

    def test_parse_failures(self) -> None:
        """Test parse failures."""

        with pytest.raises(
            PackageFileNotValid,
            match="`packages.json` file not valid, no development packages found",
        ):
            PackageManagerV1.from_json(packages={})

        with pytest.raises(
            PackageFileNotValid,
            match="`packages.json` file not valid, no third party packages found",
        ):
            PackageManagerV1.from_json(packages={"dev": {}})


class TestPackageManagerV1Sync(TestPackageManagerV1):
    """Test sync."""

    def test_sync(
        self,
    ) -> None:
        """Test sync."""

        pm = PackageManagerV1(
            path=self.packages_dir_path,
            dev_packages=OrderedDict({DUMMY_PACKAGE_ID: DUMMY_PACKAGE_HASH}),
        )

        with pytest.raises(
            ValueError,
            match=re.escape(
                "Both `update_packages` and `update_hashes` cannot be set to `True`."
            ),
        ):
            pm.sync(update_hashes=True, update_packages=True)

        with mock.patch.object(pm, "add_package") as update_patch:
            pm.sync(dev=True, third_party=False)
            update_patch.assert_called_with(
                package_id=DUMMY_PACKAGE_ID.with_hash(
                    "bafybei0000000000000000000000000000000000000000000000000000"
                ),
                with_dependencies=True,
            )

        # test package already exists.
        with mock.patch.object(pm, "add_package") as update_patch, mock.patch(
            "pathlib.Path.exists", return_value=True
        ), mock.patch(
            "aea.helpers.ipfs.base.IPFSHashOnly.get", return_value=DUMMY_PACKAGE_HASH
        ):
            pm.sync(dev=True, third_party=False)
            update_patch.assert_not_called()

        # test package already exists but hash differs.
        with mock.patch.object(pm, "add_package") as add_patch, mock.patch(
            "pathlib.Path.exists", return_value=True
        ), mock.patch(
            "aea.helpers.ipfs.base.IPFSHashOnly.get",
            return_value=DUMMY_PACKAGE_HASH + "X",
        ):
            with pytest.raises(PackageHashDoesNotMatch):
                pm.sync(dev=True, third_party=False)
            add_patch.assert_not_called()

        # update hashes
        pm = PackageManagerV1(
            path=self.packages_dir_path,
            dev_packages=OrderedDict({DUMMY_PACKAGE_ID: DUMMY_PACKAGE_HASH}),
        )

        with mock.patch.object(pm, "add_package") as add_patch, mock.patch(
            "pathlib.Path.exists", return_value=True
        ), mock.patch(
            "aea.helpers.ipfs.base.IPFSHashOnly.get",
            return_value=DUMMY_PACKAGE_HASH + "X",
        ):
            pm.sync(dev=True, third_party=False, update_hashes=True)
            add_patch.assert_not_called()

        # update packages
        pm = PackageManagerV1(
            path=self.packages_dir_path,
            dev_packages=OrderedDict({DUMMY_PACKAGE_ID: DUMMY_PACKAGE_HASH}),
        )

        with mock.patch.object(pm, "add_package") as add_patch, mock.patch.object(
            pm, "_update_package"
        ) as update_patch, mock.patch(
            "pathlib.Path.exists", return_value=True
        ), mock.patch(
            "aea.helpers.ipfs.base.IPFSHashOnly.get",
            return_value=DUMMY_PACKAGE_HASH + "X",
        ):
            pm.sync(dev=True, third_party=False, update_packages=True)
            update_patch.assert_called_once()

        # third party
        pm = PackageManagerV1(
            path=self.packages_dir_path,
            third_party_packages=OrderedDict({DUMMY_PACKAGE_ID: DUMMY_PACKAGE_HASH}),
        )

        with mock.patch.object(pm, "add_package") as update_patch:
            pm.sync(dev=False, third_party=True)
            update_patch.assert_called_with(
                package_id=DUMMY_PACKAGE_ID.with_hash(
                    "bafybei0000000000000000000000000000000000000000000000000000"
                ),
                with_dependencies=True,
            )

        # 3d part hashes
        pm = PackageManagerV1(
            path=self.packages_dir_path,
            third_party_packages=OrderedDict({DUMMY_PACKAGE_ID: DUMMY_PACKAGE_HASH}),
        )

        with mock.patch.object(pm, "add_package") as update_patch:
            pm.sync(dev=False, third_party=True, update_hashes=True)
            update_patch.assert_called_with(
                package_id=DUMMY_PACKAGE_ID.with_hash(
                    "bafybei0000000000000000000000000000000000000000000000000000"
                ),
                with_dependencies=True,
            )

        # 3d part packages
        with mock.patch.object(pm, "update_package") as update_patch, mock.patch.object(
            pm,
            "_sync",
            return_value=[True, None, {DUMMY_PACKAGE_ID: DUMMY_PACKAGE_HASH}],
        ):
            pm.sync(dev=False, third_party=True, update_packages=True)
            update_patch.assert_called_with(package_id=DUMMY_PACKAGE_ID)


class TestPackageManagerV1SourceSync(TestPackageManagerV1):
    """Test sync."""

    sources = [
        "author/repo",
    ]
    dummy_dev_package_hash = (
        "bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhnx"
    )
    dummy_third_party_package_hash = (
        "bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
    )
    dummy_third_party_package_latest = (
        "bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhnz"
    )
    dummy_dev_package = PackageId(
        package_type=PackageType.PROTOCOL,
        public_id=PublicId(author="author", name="dev_package"),
    )
    dummy_third_party_package = PackageId(
        package_type=PackageType.PROTOCOL,
        public_id=PublicId(author="author", name="third_party_package"),
    )
    dummy_packages = {
        "dev": {dummy_dev_package.to_uri_path: dummy_dev_package_hash},
        "third_party": {
            dummy_third_party_package.to_uri_path: dummy_third_party_package_hash
        },
    }

    def test_source_sync(
        self,
    ) -> None:
        """Test source sync."""

        latest_packages = copy.deepcopy(self.dummy_packages)
        latest_packages["dev"][
            self.dummy_third_party_package.to_uri_path
        ] = self.dummy_third_party_package_latest

        pm = PackageManagerV1.from_json(packages=self.dummy_packages)
        with mock.patch.object(pm, "_get_latest_tag"), mock.patch.object(
            pm, "_get_packages_json", return_value=latest_packages
        ):
            pm.sync(third_party=False, sources=self.sources)

        assert (
            pm.get_package_hash(self.dummy_third_party_package)
            == self.dummy_third_party_package_latest
        )

    def test_source_sync_failures(
        self,
    ) -> None:
        """Test source sync failures."""

        pm = PackageManagerV1.from_json(packages=self.dummy_packages)
        with pytest.raises(
            PackagesSourceNotValid, match="Provided source name is not valid `source`"
        ):
            pm.sync(
                third_party=False,
                sources=[
                    "source",
                ],
            )

        with pytest.raises(
            PackagesSourceNotValid,
            match="Fetching tags from `author/repo` failed with message 'Not Found'",
        ), mock.patch(
            "aea.package_manager.v1.r_get",
            return_value=mock.MagicMock(
                status_code=404, json=lambda: {"message": "Not Found"}
            ),
        ):
            pm.sync(
                third_party=False,
                sources=self.sources,
            )

        with pytest.raises(
            PackagesSourceNotValid,
            match="Fetching packages from `author/repo` failed with message '404: Not Found'",
        ), mock.patch(
            "aea.package_manager.v1.r_get",
            return_value=mock.MagicMock(
                status_code=404,
                text="404: Not Found",
            ),
        ):
            with mock.patch.object(pm, "_get_latest_tag", return_value="latest"):
                pm.sync(
                    third_party=False,
                    sources=self.sources,
                )


class TestPackageManagerV1UpdateFingerprint(TestPackageManagerV1):
    """Test update fingerprints."""

    def test_update_fingerprints(self, caplog) -> None:
        """Test update fingerprints."""

        package_id = PackageId.from_uri_path("protocol/open_aea/signing/1.0.0")
        package_dir_rel = Path(
            PACKAGES,
            package_id.author,
            package_id.package_type.to_plural(),
            package_id.name,
        )
        original_package = ROOT_DIR / package_dir_rel
        package_hash = IPFSHashOnly.get(str(original_package))

        with tempfile.TemporaryDirectory() as temp_dir:
            packages_dir = Path(temp_dir, PACKAGES)
            temp_package = Path(temp_dir, *package_dir_rel.parts)

            os.makedirs(temp_package.parent)

            shutil.copytree(original_package, temp_package)
            pm = PackageManagerV1(
                path=packages_dir,
                dev_packages=OrderedDict({package_id: package_hash}),
            )

            (temp_package / "__init__.py").write_text("hello")
            with caplog.at_level(logging.ERROR):
                assert pm.verify() == 1
                assert (
                    "Fingerprints does not match for (protocol, open_aea/signing:1.0.0)"
                    in caplog.text
                )

            pm.update_package_hashes(selector_prompt=lambda: "dev")
            assert pm.verify() == 0


class TestHashUpdateDev(BaseAEATestCase):
    """Test hash update."""

    use_packages_dir: bool = True

    def test_update_package_hashes(
        self,
    ) -> None:
        """Test update package hash method."""

        packages_v1 = json.loads(PACKAGE_JSON_FILE.read_text(encoding="utf-8"))

        original_hash = packages_v1["dev"][EXAMPLE_PACKAGE_ID.to_uri_path]
        packages_v1["dev"][EXAMPLE_PACKAGE_ID.to_uri_path] = DUMMY_PACKAGE_HASH

        packages_json_file = self.packages_dir_path / "packages.json"
        packages_json_file.write_text(json.dumps(obj=packages_v1))
        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        pm.update_package_hashes(selector_prompt=lambda: "dev").dump()
        packages_v1_updated = json.loads(packages_json_file.read_text(encoding="utf-8"))

        assert pm.dev_packages[EXAMPLE_PACKAGE_ID] == original_hash
        assert (
            packages_v1_updated["dev"][EXAMPLE_PACKAGE_ID.to_uri_path] == original_hash
        )

        with mock.patch.object(
            pm,
            "iter_dependency_tree",
            return_value=[EXAMPLE_PACKAGE_ID],
        ), mock.patch.object(
            pm, "is_third_party_package", return_value=True
        ), mock.patch.object(
            pm, "is_dev_package", return_value=False
        ), mock.patch.object(
            pm,
            "calculate_hash_from_package_id",
            return_value=EXAMPLE_PACKAGE_HASH,
        ), mock.patch.object(
            pm._logger, "warning"
        ) as mock_warning:
            pm._third_party_packages[EXAMPLE_PACKAGE_ID] = EXAMPLE_PACKAGE_HASH
            pm.update_package_hashes(selector_prompt=package_type_selector_prompt)
            mock_warning.assert_not_called()

    def test_add_new_package(self) -> None:
        """Test adding new package."""
        package = PackageId(
            package_type=PackageType.SKILL,
            public_id=PublicId.from_str(
                "valory/abstract_round_abci:0.1.0:bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
            ),
        )

        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        with mock.patch.object(
            pm, "iter_dependency_tree", return_value=[package]
        ), mock.patch.object(
            pm,
            "calculate_hash_from_package_id",
            return_value=EXAMPLE_PACKAGE_HASH,
        ), mock.patch.object(
            pm,
            "update_fingerprints",
        ), mock.patch.object(
            pm,
            "update_dependencies",
        ):
            pm.update_package_hashes(selector_prompt=lambda: "dev")

        assert pm.get_package_hash(package_id=package) == EXAMPLE_PACKAGE_HASH
        assert pm.dev_packages[package.without_hash()] == EXAMPLE_PACKAGE_HASH

    def test_add_new_package_failures(self) -> None:
        """Test adding new package."""
        package = PackageId(
            package_type=PackageType.SKILL,
            public_id=PublicId.from_str(
                "valory/abstract_round_abci:0.1.0:bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
            ),
        )

        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        with mock.patch.object(
            pm, "iter_dependency_tree", return_value=[package]
        ), mock.patch.object(
            pm,
            "calculate_hash_from_package_id",
            return_value=EXAMPLE_PACKAGE_HASH,
        ), mock.patch.object(
            pm,
            "update_fingerprints",
        ), mock.patch.object(
            pm,
            "update_dependencies",
        ):
            with pytest.raises(
                PackageNotValid,
                match="Found a package which is not listed in the `packages.json`",
            ):
                pm.update_package_hashes()

    def test_skip_missing_package(self) -> None:
        """Test adding new package."""
        package = PackageId(
            package_type=PackageType.SKILL,
            public_id=PublicId.from_str(
                "valory/abstract_round_abci:0.1.0:bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
            ),
        )

        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        with mock.patch.object(
            pm, "iter_dependency_tree", return_value=[package]
        ), mock.patch.object(
            pm,
            "calculate_hash_from_package_id",
            return_value=EXAMPLE_PACKAGE_HASH,
        ), mock.patch.object(
            pm,
            "update_fingerprints",
        ), mock.patch.object(
            pm,
            "update_dependencies",
        ):
            pm.update_package_hashes(skip_missing=True)

        assert pm.get_package_hash(package_id=package) is None
        assert package.without_hash() not in pm.dev_packages


class TestHashUpdateThirdParty(BaseAEATestCase):
    """Test hash update."""

    use_packages_dir: bool = True

    def test_update_package_hashes(self, caplog) -> None:
        """Test update package hash method."""

        packages_v1 = json.loads(PACKAGE_JSON_FILE.read_text(encoding="utf-8"))
        original_hash = packages_v1["dev"].pop(EXAMPLE_PACKAGE_ID.to_uri_path)
        packages_v1["third_party"][EXAMPLE_PACKAGE_ID.to_uri_path] = DUMMY_PACKAGE_HASH

        packages_json_file = self.packages_dir_path / "packages.json"
        packages_json_file.write_text(json.dumps(obj=packages_v1))
        pm = PackageManagerV1.from_dir(self.packages_dir_path)

        with caplog.at_level(logging.WARNING):
            pm.update_package_hashes(selector_prompt=lambda: "third_party").dump()
            packages_v1_updated = json.loads(
                packages_json_file.read_text(encoding="utf-8")
            )

            assert pm.third_party_packages[EXAMPLE_PACKAGE_ID] == original_hash
            assert (
                packages_v1_updated["third_party"][EXAMPLE_PACKAGE_ID.to_uri_path]
                == original_hash
            )
            assert "Hash change detected for third party package" in caplog.text


class TestVerifyFailure(BaseAEATestCase):
    """Test verify method."""

    use_packages_dir: bool = True

    def test_verify_method(self, caplog) -> None:
        """Test update package hash method."""

        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        assert pm.verify() == 0

        with caplog.at_level(logging.ERROR), mock.patch.object(
            pm,
            "check_dependencies",
            return_value=[
                (EXAMPLE_PACKAGE_ID, DepedencyMismatchErrors.HASH_NOT_FOUND),
                (EXAMPLE_PACKAGE_ID, DepedencyMismatchErrors.HASH_DOES_NOT_MATCH),
            ],
        ):
            assert pm.verify() == 1
            assert (
                "Package contains a dependency that is not defined in the"
                in caplog.text
            )
            assert "Dependency check failed\nHash does not match for" in caplog.text

        with mock.patch.object(
            pm, "check_dependencies", side_effect=ValueError("expected")
        ), mock.patch("traceback.print_exc") as print_tb_mock:
            assert pm.verify() == 1
            print_tb_mock.assert_called_once()

        # updating the `packages/open_aea/protocols/signing/__init__.py` file
        init_file = (
            pm.package_path_from_package_id(package_id=EXAMPLE_PACKAGE_ID)
            / INIT_FILE_NAME
        )
        init_file.write_text("dummy")

        with caplog.at_level(logging.ERROR), mock.patch(
            "aea.package_manager.v1.check_fingerprint",
            return_value=True,
        ):
            pm = PackageManagerV1.from_dir(self.packages_dir_path)

            assert pm.verify() == 1
            assert f"Hash does not match for {EXAMPLE_PACKAGE_ID}" in caplog.text
            assert (
                f"Dependency check failed\nHash does not match for {EXAMPLE_PACKAGE_ID}"
                in caplog.text
            )

    def test_fingerprint_failure(self, caplog) -> None:
        """Test update package hash method."""

        pm = PackageManagerV1(
            path=self.packages_dir_path,
            dev_packages=OrderedDict(
                {
                    EXAMPLE_PACKAGE_ID: EXAMPLE_PACKAGE_HASH,
                }
            ),
        )

        with mock.patch.object(
            pm,
            "_calculate_hash",
            return_value=DUMMY_PACKAGE_HASH,
        ), mock.patch.object(
            pm,
            "is_third_party_package",
            return_value=True,
        ):
            assert pm.verify() == 1
            assert pm._third_party_packages[EXAMPLE_PACKAGE_ID] == DUMMY_PACKAGE_HASH

        with caplog.at_level(logging.ERROR), mock.patch(
            "aea.package_manager.v1.check_fingerprint",
            return_value=False,
        ), mock.patch.object(
            pm,
            "iter_dependency_tree",
            return_value=[
                EXAMPLE_PACKAGE_ID,
            ],
        ), mock.patch.object(
            pm,
            "is_third_party_package",
            return_value=True,
        ):
            assert pm.verify() == 1
            assert (
                f"Fingerprints does not match for {EXAMPLE_PACKAGE_ID}" in caplog.text
            )

    def test_missing_hash(self, caplog) -> None:
        """Test update package hash method."""

        pm = PackageManagerV1(path=self.packages_dir_path)

        with caplog.at_level(logging.ERROR), mock.patch(
            "aea.package_manager.v1.check_fingerprint",
            return_value=True,
        ), mock.patch.object(
            pm,
            "iter_dependency_tree",
            return_value=[
                EXAMPLE_PACKAGE_ID,
            ],
        ):
            assert pm.verify() == 1
            assert f"Cannot find hash for {EXAMPLE_PACKAGE_ID}" in caplog.text

    def test_add_new_package_third_party(self) -> None:
        """Test adding new package."""
        package = PackageId(
            package_type=PackageType.SKILL,
            public_id=PublicId.from_str(
                "valory/abstract_round_abci:0.1.0:bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
            ),
        )

        pm = PackageManagerV1.from_dir(self.packages_dir_path)
        with mock.patch.object(
            pm, "iter_dependency_tree", return_value=[package]
        ), mock.patch.object(
            pm,
            "calculate_hash_from_package_id",
            return_value=EXAMPLE_PACKAGE_HASH,
        ), mock.patch.object(
            pm,
            "update_fingerprints",
        ), mock.patch.object(
            pm,
            "update_dependencies",
        ):
            pm.update_package_hashes(selector_prompt=lambda: "third_party")

        assert pm.get_package_hash(package_id=package) == EXAMPLE_PACKAGE_HASH
        assert pm.third_party_packages[package.without_hash()] == EXAMPLE_PACKAGE_HASH


@pytest.mark.integration
def test_package_manager_add_item_dependency_support():
    """Check PackageManager.add_packages works with dependencies on real packages."""
    with tempfile.TemporaryDirectory() as tmpdir:
        package_manager = PackageManagerV1(Path(tmpdir))
        package_manager.add_package(TEST_SKILL_ID)
        assert len(package_manager.third_party_packages) == 1
        package_manager.dump()

    with tempfile.TemporaryDirectory() as tmpdir:
        package_manager = PackageManagerV1(Path(tmpdir))
        package_manager.add_package(
            TEST_SKILL_ID,
            with_dependencies=True,
        )
        assert len(package_manager.third_party_packages) > 1

        str_packages = str(package_manager.third_party_packages)
        # check some  deps
        assert "protocol, valory/abci" in str_packages
        assert "skill, valory/abstract_round_abci" in str_packages
        assert "protocol, valory/tendermint" in str_packages
        package_manager.dump()


@mock.patch("aea.package_manager.base.load_fetch_ipfs")
def test_package_manager_add_item_dependency_support_mock(fetch_mock):
    """Check PackageManager.add_packages works with dependencies on mocks."""
    FAKE_PACKAGES = [
        PackageId(
            package_type=PackageType.SKILL,
            public_id=PublicId.from_str(
                f"valory/abstract_{i}:0.1.0:bafybeifh4qtjurq5637ykxexzexca5l4n6t4ujw26tpnern2swajanvhny"
            ),
        )
        for i in range(3)
    ]

    # no deps
    with tempfile.TemporaryDirectory() as tmpdir:
        package_manager = PackageManagerV1(Path(tmpdir))
        with mock.patch.object(package_manager, "calculate_hash_from_package_id"):
            package_manager.add_package(TEST_SKILL_ID)
            assert len(package_manager.third_party_packages) == 1

    # deps loading
    with tempfile.TemporaryDirectory() as tmpdir:
        package_manager = PackageManagerV1(Path(tmpdir))
        with mock.patch.object(
            package_manager, "calculate_hash_from_package_id"
        ), mock.patch.object(
            package_manager,
            "get_package_dependencies",
            side_effect=[
                [FAKE_PACKAGES[0]],
                [FAKE_PACKAGES[1]],
                [FAKE_PACKAGES[2]],
                [],
                [],
                [],
                [],
            ],
        ):
            package_manager.add_package(
                TEST_SKILL_ID,
                with_dependencies=True,
            )
            assert len(package_manager.third_party_packages) == 4


@mock.patch("aea.package_manager.base.load_fetch_ipfs")
def test_package_manager_add_package_already_installed(fetch_mock: mock.Mock):
    """Test package already installed."""
    # version already installed
    with tempfile.TemporaryDirectory() as tmpdir:
        package_manager = PackageManagerV1(Path(tmpdir))
        with mock.patch.object(
            package_manager, "get_package_version_with_hash", return_value=TEST_SKILL_ID
        ), mock.patch.object(
            package_manager, "is_package_files_exist", return_value=True
        ), mock.patch.object(
            package_manager, "calculate_hash_from_package_id"
        ):
            package_manager.add_package(TEST_SKILL_ID)
            fetch_mock.assert_not_called()


@mock.patch("aea.package_manager.base.load_fetch_ipfs")
def test_package_manager_add_package_can_be_updated(fetch_mock: mock.Mock):
    """Test package update on add_package."""
    # version already installed
    data = TEST_SKILL_ID.public_id.json
    data["version"] = "0.0.1"
    OLDER_VERSION = PackageId(
        package_type=TEST_SKILL_ID.package_type, public_id=PublicId.from_json(data)
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        package_manager = PackageManagerV1(Path(tmpdir))
        with mock.patch.object(
            package_manager, "get_package_version_with_hash", return_value=OLDER_VERSION
        ), mock.patch.object(
            package_manager, "is_package_files_exist", return_value=True
        ), mock.patch.object(
            package_manager, "calculate_hash_from_package_id"
        ), mock.patch.object(
            package_manager, "_remove_package_dir"
        ) as remove_mock:
            with pytest.raises(
                ValueError,
                match="Required package and package in the registry does not match",
            ):
                package_manager.add_package(TEST_SKILL_ID)
            fetch_mock.assert_not_called()

            package_manager.add_package(TEST_SKILL_ID, allow_update=True)
            remove_mock.assert_called_once_with(TEST_SKILL_ID)


class TestGetPackageDependencies(BaseAEATestCase):
    """Test test_get_package_dependencies."""

    HTTP_CONNECTION_PACKAGE = PackageId(
        package_type=PackageType.CONNECTION,
        public_id=PublicId(
            author="valory", name="http_client", version="1.0.0"
        ).to_latest(),
    )
    use_packages_dir: bool = True

    def test_get_package_dependencies(
        self,
    ) -> None:
        """Test get_package_dependencies method."""

        pm = PackageManagerV1.from_dir(self.t / self.packages_dir_path)
        assert "PackageId(protocol, valory/http" in str(
            pm.get_package_dependencies(self.HTTP_CONNECTION_PACKAGE)
        )


class TestGetPackageVersionWithHash(BaseAEATestCase):
    """Test get_package_version_with_hash."""

    HTTP_CONNECTION_PACKAGE = PackageId(
        package_type=PackageType.CONNECTION,
        public_id=PublicId(
            author="valory", name="http_client", version="1.0.0"
        ).to_latest(),
    )
    use_packages_dir: bool = True

    def test_get_package_version_with_hash(
        self,
    ) -> None:
        """Test get_package_dependencies method."""

        pm = PackageManagerV1.from_dir(self.t / self.packages_dir_path)
        package_id = pm.get_package_version_with_hash(self.HTTP_CONNECTION_PACKAGE)
        assert package_id.package_hash



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the component loading utils."""


import re
import sys
from pathlib import Path
from typing import Dict, List

from aea.components.base import perform_load_aea_package
from aea.configurations.constants import CONNECTIONS, CONTRACTS, PROTOCOLS, SKILLS


PACKAGES_RE = re.compile(r"^packages\.(\w+)\.(\w+)\.(\w+)$", re.I)


def _enlist_component_packages() -> Dict[str, List[Dict[str, str]]]:
    """List all components packages already loaded."""
    result: Dict[str, List[Dict[str, str]]] = {}

    for name, mod in sys.modules.items():
        match = PACKAGES_RE.match(name)
        if not match:
            continue
        author, package_type, package_name = match.groups()
        packages = result.get(package_type, [])
        package_data = {
            "author": author,
            "package_type": package_type,
            "package_name": package_name,
            "dir": mod.__dict__["__path__"][0],
        }
        packages.append(package_data)
        result[package_type] = packages
    return result


# no cover cause executed in a subprocess, not possible to track
def _populate_packages(packages: dict) -> None:  # pragma: nocover
    """Load packages as python modules."""
    for package_type in [PROTOCOLS, CONTRACTS, CONNECTIONS, SKILLS]:
        for package in packages.get(package_type, []):
            # load package
            print(11, package)
            perform_load_aea_package(
                dir_=Path(package["dir"]),
                author=package["author"],
                package_type_plural=package["package_type"],
                package_name=package["package_name"],
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the task class of the gym skill."""
# pylint: skip-file

from unittest.mock import patch

from packages.fetchai.skills.gym.tests.intermediate_class import GymTestCase


class TestTask(GymTestCase):
    """Test Task of gym."""

    def test__init__(self):
        """Test the __init__ method of the GymTask class."""
        assert self.task.nb_steps == self.nb_steps
        assert self.task.is_rl_agent_training is False

    def test_properties(self):
        """Test the properties of the GymTask class."""
        assert self.task.proxy_env == self.task._proxy_env
        assert self.task.proxy_env_queue == self.task._proxy_env.queue

    def test_setup(self):
        """Test the setup method of the GymTask class."""
        # operation
        with patch.object(self.logger, "info") as mock_logger:
            self.task.setup()

        # after
        mock_logger.assert_any_call("Gym task: setup method called.")

    def test_execute_i(self):
        """Test the execute method of the GymTask class where agent is NOT trained/training."""
        # before
        self.task.proxy_env._is_rl_agent_trained = False
        self.task.is_rl_agent_training = False

        # operation
        with patch.object(self.task._rl_agent, "fit") as mock_fit:
            with patch.object(self.logger, "info") as mock_logger:
                self.task.execute()

        # after
        # _start_training
        mock_logger.assert_any_call("Training starting ...")
        assert self.task.is_rl_agent_training is True

        # _fit
        mock_fit.assert_called_with(self.task.proxy_env, self.nb_steps)
        mock_logger.assert_any_call("Training finished. You can exit now via CTRL+C.")

    def test_execute_ii(self):
        """Test the execute method of the GymTask class where agent IS trained/training."""
        # before
        self.task.proxy_env._is_rl_agent_trained = True
        self.task.is_rl_agent_training = True

        # operation
        with patch.object(self.task.proxy_env, "close") as mock_close:
            self.task.execute()

        # after
        # _stop_training
        assert self.task.is_rl_agent_training is False
        mock_close.assert_called_once()

    def test_teardown(self):
        """Test the teardown method of the GymTask class."""
        self.task.is_rl_agent_training = True

        # operation
        with patch.object(self.task.proxy_env, "close") as mock_close:
            with patch.object(self.logger, "info") as mock_logger:
                self.task.teardown()

        # after
        mock_logger.assert_any_call("Gym Task: teardown method called.")

        # _stop_training
        assert self.task.is_rl_agent_training is False
        mock_close.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test confiugurations for the package."""
# pylint: skip-file

import logging
from pathlib import Path
from typing import Any, AsyncGenerator, Dict, Generator, cast
from unittest.mock import MagicMock

import pytest
import pytest_asyncio
from aea_ledger_ethereum import EthereumCrypto
from aea_ledger_ethereum.test_tools.constants import (
    ETHEREUM_TESTNET_CONFIG as _DEFAULT_ETHEREUM_TESTNET_CONFIG,
)
from aea_ledger_ethereum.test_tools.fixture_helpers import (
    DEFAULT_GANACHE_ADDR,
    DEFAULT_GANACHE_PORT,
)

from aea.configurations.base import ComponentType, ContractConfig
from aea.configurations.constants import DEFAULT_LEDGER
from aea.configurations.loader import load_component_configuration
from aea.connections.base import Connection
from aea.contracts.base import Contract, contract_registry
from aea.crypto.ledger_apis import DEFAULT_LEDGER_CONFIGS, LedgerApi
from aea.crypto.registries import ledger_apis_registry, make_crypto
from aea.crypto.wallet import CryptoStore
from aea.identity.base import Identity


PACKAGE_DIR = Path(__file__).parent.parent

DEFAULT_ETHEREUM_TESTNET_CONFIG = {
    **_DEFAULT_ETHEREUM_TESTNET_CONFIG,
    "default_gas_price_strategy": "eip1559",
}


def get_register_contract(directory: Path) -> Contract:
    """Get and register the erc1155 contract package."""
    configuration = load_component_configuration(ComponentType.CONTRACT, directory)
    configuration._directory = directory  # pylint: disable=protected-access
    configuration = cast(ContractConfig, configuration)

    if str(configuration.public_id) not in contract_registry.specs:
        # load contract into sys modules
        Contract.from_config(configuration)

    contract = contract_registry.make(str(configuration.public_id))
    return contract


@pytest.fixture()
def ganache_addr() -> str:
    """Get the ganache addr"""
    return DEFAULT_GANACHE_ADDR


@pytest.fixture()
def ganache_port() -> int:
    """Get the ganache port"""
    return DEFAULT_GANACHE_PORT


@pytest.fixture(scope="function")
def ethereum_testnet_config(ganache_addr: str, ganache_port: int) -> Dict:
    """Get Ethereum ledger api configurations using Ganache."""
    new_uri = f"{ganache_addr}:{ganache_port}"
    new_config = DEFAULT_ETHEREUM_TESTNET_CONFIG.copy()
    new_config["address"] = new_uri
    return new_config


@pytest.fixture(scope="function")
def update_default_ethereum_ledger_api(ethereum_testnet_config: Dict) -> Generator:
    """Change temporarily default Ethereum ledger api configurations to interact with local Ganache."""
    old_config = DEFAULT_LEDGER_CONFIGS.pop(EthereumCrypto.identifier)
    DEFAULT_LEDGER_CONFIGS[EthereumCrypto.identifier] = ethereum_testnet_config
    yield
    DEFAULT_LEDGER_CONFIGS.pop(EthereumCrypto.identifier)
    DEFAULT_LEDGER_CONFIGS[EthereumCrypto.identifier] = old_config


def make_ledger_api_connection(
    ethereum_testnet_config: Dict = DEFAULT_ETHEREUM_TESTNET_CONFIG,
) -> Connection:
    """Make a connection."""
    crypto = make_crypto(DEFAULT_LEDGER)
    identity = Identity("name", crypto.address, crypto.public_key)
    crypto_store = CryptoStore()
    directory = PACKAGE_DIR
    connection = Connection.from_dir(
        str(directory),
        data_dir=MagicMock(),
        identity=identity,
        crypto_store=crypto_store,
    )
    connection = cast(Connection, connection)
    connection._logger = logging.getLogger("packages.valory.connections.ledger")

    # use testnet config
    connection.configuration.config.get("ledger_apis", {})[
        "ethereum"
    ] = ethereum_testnet_config

    connection.request_retry_attempts = 1  # type: ignore
    connection.request_retry_attempts = 2  # type: ignore
    return connection


@pytest_asyncio.fixture()
async def ledger_apis_connection(
    request: Any, ethereum_testnet_config: Dict
) -> AsyncGenerator:
    """Make a connection."""
    connection = make_ledger_api_connection(ethereum_testnet_config)
    await connection.connect()
    yield connection
    await connection.disconnect()


@pytest.fixture()
def ledger_api(
    ethereum_testnet_config: Dict,
) -> Generator[LedgerApi, None, None]:
    """Ledger api fixture."""
    ledger_id, config = EthereumCrypto.identifier, ethereum_testnet_config
    api = ledger_apis_registry.make(ledger_id, **config)
    yield api



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for the helpers/profiling module."""
import platform
import re
from typing import Dict, List, Optional

import pytest

from aea.helpers.profiling import Profiling
from aea.protocols.base import Message

from tests.common.utils import wait_for_condition


if platform.system() == "Windows":  # pragma: nocover
    # type: ignore  # pylint: disable=import-error,import-outside-toplevel,unsed-import
    import win32timezone  # type: ignore

    _ = win32timezone

    import win32process  # type: ignore  # pylint: disable=import-error,import-outside-toplevel,unsed-import # noqa: F401


TIMEOUT = 20
MESSAGE_NUMBER = 10
DUMMIES_NUMBER = 1000


class DummyClass:
    """Dummy class for counting purposes"""


class MessageContainer:
    """Dummy class for counting purposes"""

    def __init__(self, other: Optional["MessageContainer"] = None) -> None:
        """Initializer"""
        self.messages: List[Message] = other.messages if other else create_messages()


def __create_two_return_one() -> MessageContainer:
    """Create two message containers return only one referencing the same messages"""
    # create MessageContainer twice, Messages in them only once (shared)
    # only one of them is tracked by garbage collector, as the inner is unbound
    return MessageContainer(MessageContainer())


def create_dummies() -> List[DummyClass]:
    """Create n dummy class instances"""
    return [DummyClass() for _ in range(DUMMIES_NUMBER)]


def create_messages() -> List[Message]:
    """Create n messages"""
    return [Message() for _ in range(MESSAGE_NUMBER)]


def extract_object_counts(log: str) -> Dict[str, Dict[str, int]]:
    """Extract object counts from the profiling log."""
    result: Dict[str, Dict[str, int]] = {"created": {}, "present": {}, "gc": {}}

    for line in log.split("\n"):
        # Created (tracked objects)
        match = re.match(r".*\* (?P<field>.*) \(present\):  (?P<count>\d+).*", line)
        if match:
            result["present"][match.groupdict()["field"]] = int(
                match.groupdict()["count"]
            )
            continue

        # Present (tracked objects)
        match = re.match(r".*\* (?P<field>.*) \(created\):  (?P<count>\d+).*", line)
        if match:
            result["created"][match.groupdict()["field"]] = int(
                match.groupdict()["count"]
            )
            continue

        # Present (garbage collector)
        match = re.match(r".*\* (?P<field>.*) \(gc\):  (?P<count>\d+).*", line)
        if match:
            result["gc"][match.groupdict()["field"]] = int(match.groupdict()["count"])

    return result


def test_basic_profiling():
    """Test profiling tool."""

    def output_function(report):
        """Test output function"""
        nonlocal result
        result = report

    result, types_to_track = "", [Message]
    p = Profiling(types_to_track, 1, output_function=output_function)
    p.start()
    m = Message()
    del m
    wait_for_condition(lambda: p.is_running, timeout=TIMEOUT)

    try:
        wait_for_condition(lambda: result, timeout=TIMEOUT)
        assert "Profiling details" in result
        assert p.object_counts[Message] == [1, 1]
    finally:
        p.stop()
        p.wait_completed(sync=True, timeout=TIMEOUT)


@pytest.mark.profiling
def test_profiling_instance_number():
    """Test profiling tool."""

    def output_function(report):
        """Test output function"""
        nonlocal result
        result = report

    result, types_to_track = "", [Message]
    p = Profiling(types_to_track, 1, output_function=output_function)
    p.start()
    wait_for_condition(lambda: p.is_running, timeout=TIMEOUT)

    __reference, messages = create_dummies(), create_messages()  # noqa: F841

    try:
        wait_for_condition(lambda: result, timeout=TIMEOUT)
        count_dict = extract_object_counts(result)
        assert count_dict["created"] == {"Message": MESSAGE_NUMBER}
        assert count_dict["present"] == {"Message": MESSAGE_NUMBER}
        assert count_dict["gc"]["DummyClass"] == DUMMIES_NUMBER
        # create no new, cut existing messages by half
        result, messages = "", messages[: MESSAGE_NUMBER // 2]
        wait_for_condition(lambda: result, timeout=TIMEOUT)
        count_dict = extract_object_counts(result)
        assert count_dict["created"] == {"Message": MESSAGE_NUMBER}
        assert count_dict["present"] == {"Message": MESSAGE_NUMBER // 2}
        # create a second series of messages, add to the remaining half
        result, messages = "", [messages + create_messages()]
        wait_for_condition(lambda: result, timeout=TIMEOUT)
        count_dict = extract_object_counts(result)
        assert count_dict["created"] == {"Message": 2 * MESSAGE_NUMBER}
        assert count_dict["present"] == {"Message": 1.5 * MESSAGE_NUMBER}
    finally:
        p.stop()
        p.wait_completed(sync=True, timeout=TIMEOUT)


@pytest.mark.profiling
def test_profiling_cross_reference():
    """Test profiling tool."""

    def output_function(report):
        """Test output function"""
        nonlocal result
        result = report

    result, types_to_track = "", [Message, MessageContainer]
    p = Profiling(types_to_track, 1, output_function=output_function)
    p.start()
    wait_for_condition(lambda: p.is_running, timeout=TIMEOUT)

    __reference = __create_two_return_one()  # noqa: F841
    expected_created = {"Message": MESSAGE_NUMBER, "MessageContainer": 2}
    expected_present = {"Message": MESSAGE_NUMBER, "MessageContainer": 1}

    try:
        wait_for_condition(lambda: result, timeout=TIMEOUT)
        count_dict = extract_object_counts(result)
        assert count_dict["created"] == expected_created
        assert count_dict["present"] == expected_present
    finally:
        p.stop()
        p.wait_completed(sync=True, timeout=TIMEOUT)


@pytest.mark.profiling
def test_profiling_counts_not_equal():
    """Test profiling tool."""

    def output_function(report):
        """Test output function"""
        nonlocal result
        result = report

    result, types_to_track = "", [Message, MessageContainer, DummyClass]
    p = Profiling(types_to_track, 1, output_function=output_function)
    p.start()
    wait_for_condition(lambda: p.is_running, timeout=TIMEOUT)

    __reference = create_dummies(), __create_two_return_one()  # noqa: F841
    expected_shared = {"Message": MESSAGE_NUMBER, "DummyClass": DUMMIES_NUMBER}
    expected_created = {**expected_shared, "MessageContainer": 2}
    expected_present = {**expected_shared, "MessageContainer": 1}

    try:
        wait_for_condition(lambda: result, timeout=TIMEOUT)
        count_dict = extract_object_counts(result)
        assert count_dict["created"] == expected_created
        assert count_dict["present"] == expected_present
        assert count_dict["gc"].get("DummyClass", 0) == DUMMIES_NUMBER
        assert "Message" not in count_dict["gc"]
        assert "MessageContainer" not in count_dict["gc"]
    finally:
        p.stop()
        p.wait_completed(sync=True, timeout=TIMEOUT)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: models.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cmodels.proto\x12\x19\x61\x65\x61.helpers.search.models\"\xc1\x19\n\x05Query\x1a\xc0\x01\n\tAttribute\x12\x0c\n\x04name\x18\x01 \x01(\t\x12=\n\x04type\x18\x02 \x01(\x0e\x32/.aea.helpers.search.models.Query.Attribute.Type\x12\x10\n\x08required\x18\x03 \x01(\x08\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\"?\n\x04Type\x12\n\n\x06\x44OUBLE\x10\x00\x12\x07\n\x03INT\x10\x01\x12\x08\n\x04\x42OOL\x10\x02\x12\n\n\x06STRING\x10\x03\x12\x0c\n\x08LOCATION\x10\x04\x1an\n\tDataModel\x12\x0c\n\x04name\x18\x01 \x01(\t\x12>\n\nattributes\x18\x02 \x03(\x0b\x32*.aea.helpers.search.models.Query.Attribute\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x1a$\n\x08Location\x12\x0b\n\x03lon\x18\x01 \x01(\x01\x12\x0b\n\x03lat\x18\x02 \x01(\x01\x1a\x99\x01\n\x05Value\x12\x10\n\x06string\x18\x01 \x01(\tH\x00\x12\x10\n\x06\x64ouble\x18\x02 \x01(\x01H\x00\x12\x11\n\x07\x62oolean\x18\x03 \x01(\x08H\x00\x12\x11\n\x07integer\x18\x04 \x01(\x03H\x00\x12=\n\x08location\x18\x05 \x01(\x0b\x32).aea.helpers.search.models.Query.LocationH\x00\x42\x07\n\x05value\x1aN\n\x08KeyValue\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32&.aea.helpers.search.models.Query.Value\x1a\x80\x01\n\x08Instance\x12\x39\n\x05model\x18\x01 \x01(\x0b\x32*.aea.helpers.search.models.Query.DataModel\x12\x39\n\x06values\x18\x02 \x03(\x0b\x32).aea.helpers.search.models.Query.KeyValue\x1a+\n\nStringPair\x12\r\n\x05\x66irst\x18\x01 \x01(\t\x12\x0e\n\x06second\x18\x02 \x01(\t\x1a(\n\x07IntPair\x12\r\n\x05\x66irst\x18\x01 \x01(\x03\x12\x0e\n\x06second\x18\x02 \x01(\x03\x1a+\n\nDoublePair\x12\r\n\x05\x66irst\x18\x01 \x01(\x01\x12\x0e\n\x06second\x18\x02 \x01(\x01\x1a\x83\x01\n\x0cLocationPair\x12\x38\n\x05\x66irst\x18\x01 \x01(\x0b\x32).aea.helpers.search.models.Query.Location\x12\x39\n\x06second\x18\x02 \x01(\x0b\x32).aea.helpers.search.models.Query.Location\x1a\xa1\x02\n\x05Range\x12\x42\n\x0bstring_pair\x18\x01 \x01(\x0b\x32+.aea.helpers.search.models.Query.StringPairH\x00\x12@\n\x0cinteger_pair\x18\x02 \x01(\x0b\x32(.aea.helpers.search.models.Query.IntPairH\x00\x12\x42\n\x0b\x64ouble_pair\x18\x03 \x01(\x0b\x32+.aea.helpers.search.models.Query.DoublePairH\x00\x12\x46\n\rlocation_pair\x18\x04 \x01(\x0b\x32-.aea.helpers.search.models.Query.LocationPairH\x00\x42\x06\n\x04pair\x1aW\n\x08\x44istance\x12\x39\n\x06\x63\x65nter\x18\x01 \x01(\x0b\x32).aea.helpers.search.models.Query.Location\x12\x10\n\x08\x64istance\x18\x02 \x01(\x01\x1a\xca\x01\n\x08Relation\x12\x44\n\x08operator\x18\x01 \x01(\x0e\x32\x32.aea.helpers.search.models.Query.Relation.Operator\x12\x35\n\x05value\x18\x02 \x01(\x0b\x32&.aea.helpers.search.models.Query.Value\"A\n\x08Operator\x12\x06\n\x02\x45Q\x10\x00\x12\x06\n\x02LT\x10\x01\x12\x08\n\x04LTEQ\x10\x02\x12\x06\n\x02GT\x10\x03\x12\x08\n\x04GTEQ\x10\x04\x12\t\n\x05NOTEQ\x10\x05\x1a\xca\x05\n\x03Set\x12?\n\x08operator\x18\x01 \x01(\x0e\x32-.aea.helpers.search.models.Query.Set.Operator\x12;\n\x06values\x18\x02 \x01(\x0b\x32+.aea.helpers.search.models.Query.Set.Values\x1a\xa5\x04\n\x06Values\x12\x45\n\x06string\x18\x01 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.Set.Values.StringsH\x00\x12\x45\n\x06\x64ouble\x18\x02 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.Set.Values.DoublesH\x00\x12\x44\n\x07\x62oolean\x18\x03 \x01(\x0b\x32\x31.aea.helpers.search.models.Query.Set.Values.BoolsH\x00\x12\x43\n\x07integer\x18\x04 \x01(\x0b\x32\x30.aea.helpers.search.models.Query.Set.Values.IntsH\x00\x12I\n\x08location\x18\x05 \x01(\x0b\x32\x35.aea.helpers.search.models.Query.Set.Values.LocationsH\x00\x1a\x16\n\x04Ints\x12\x0e\n\x06values\x18\x01 \x03(\x03\x1a\x19\n\x07\x44oubles\x12\x0e\n\x06values\x18\x01 \x03(\x01\x1a\x19\n\x07Strings\x12\x0e\n\x06values\x18\x01 \x03(\t\x1a\x17\n\x05\x42ools\x12\x0e\n\x06values\x18\x01 \x03(\x08\x1a\x46\n\tLocations\x12\x39\n\x06values\x18\x01 \x03(\x0b\x32).aea.helpers.search.models.Query.LocationB\x08\n\x06values\"\x1d\n\x08Operator\x12\x06\n\x02IN\x10\x00\x12\t\n\x05NOTIN\x10\x01\x1a\xc3\x06\n\x0e\x43onstraintExpr\x12\x41\n\x03or_\x18\x01 \x01(\x0b\x32\x32.aea.helpers.search.models.Query.ConstraintExpr.OrH\x00\x12\x43\n\x04\x61nd_\x18\x02 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.ConstraintExpr.AndH\x00\x12\x43\n\x04not_\x18\x03 \x01(\x0b\x32\x33.aea.helpers.search.models.Query.ConstraintExpr.NotH\x00\x12P\n\nconstraint\x18\x04 \x01(\x0b\x32:.aea.helpers.search.models.Query.ConstraintExpr.ConstraintH\x00\x1aI\n\x02Or\x12\x43\n\nexpression\x18\x01 \x03(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x1aJ\n\x03\x41nd\x12\x43\n\nexpression\x18\x01 \x03(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x1aJ\n\x03Not\x12\x43\n\nexpression\x18\x01 \x01(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x1a\xa0\x02\n\nConstraint\x12\x16\n\x0e\x61ttribute_name\x18\x01 \x01(\t\x12\x34\n\x04set_\x18\x02 \x01(\x0b\x32$.aea.helpers.search.models.Query.SetH\x00\x12\x38\n\x06range_\x18\x03 \x01(\x0b\x32&.aea.helpers.search.models.Query.RangeH\x00\x12=\n\x08relation\x18\x04 \x01(\x0b\x32).aea.helpers.search.models.Query.RelationH\x00\x12=\n\x08\x64istance\x18\x05 \x01(\x0b\x32).aea.helpers.search.models.Query.DistanceH\x00\x42\x0c\n\nconstraintB\x0c\n\nexpression\x1a\x88\x01\n\x05Model\x12\x44\n\x0b\x63onstraints\x18\x01 \x03(\x0b\x32/.aea.helpers.search.models.Query.ConstraintExpr\x12\x39\n\x05model\x18\x02 \x01(\x0b\x32*.aea.helpers.search.models.Query.DataModelB\x02H\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'models_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'H\001'
  _globals['_QUERY']._serialized_start=44
  _globals['_QUERY']._serialized_end=3309
  _globals['_QUERY_ATTRIBUTE']._serialized_start=54
  _globals['_QUERY_ATTRIBUTE']._serialized_end=246
  _globals['_QUERY_ATTRIBUTE_TYPE']._serialized_start=183
  _globals['_QUERY_ATTRIBUTE_TYPE']._serialized_end=246
  _globals['_QUERY_DATAMODEL']._serialized_start=248
  _globals['_QUERY_DATAMODEL']._serialized_end=358
  _globals['_QUERY_LOCATION']._serialized_start=360
  _globals['_QUERY_LOCATION']._serialized_end=396
  _globals['_QUERY_VALUE']._serialized_start=399
  _globals['_QUERY_VALUE']._serialized_end=552
  _globals['_QUERY_KEYVALUE']._serialized_start=554
  _globals['_QUERY_KEYVALUE']._serialized_end=632
  _globals['_QUERY_INSTANCE']._serialized_start=635
  _globals['_QUERY_INSTANCE']._serialized_end=763
  _globals['_QUERY_STRINGPAIR']._serialized_start=765
  _globals['_QUERY_STRINGPAIR']._serialized_end=808
  _globals['_QUERY_INTPAIR']._serialized_start=810
  _globals['_QUERY_INTPAIR']._serialized_end=850
  _globals['_QUERY_DOUBLEPAIR']._serialized_start=852
  _globals['_QUERY_DOUBLEPAIR']._serialized_end=895
  _globals['_QUERY_LOCATIONPAIR']._serialized_start=898
  _globals['_QUERY_LOCATIONPAIR']._serialized_end=1029
  _globals['_QUERY_RANGE']._serialized_start=1032
  _globals['_QUERY_RANGE']._serialized_end=1321
  _globals['_QUERY_DISTANCE']._serialized_start=1323
  _globals['_QUERY_DISTANCE']._serialized_end=1410
  _globals['_QUERY_RELATION']._serialized_start=1413
  _globals['_QUERY_RELATION']._serialized_end=1615
  _globals['_QUERY_RELATION_OPERATOR']._serialized_start=1550
  _globals['_QUERY_RELATION_OPERATOR']._serialized_end=1615
  _globals['_QUERY_SET']._serialized_start=1618
  _globals['_QUERY_SET']._serialized_end=2332
  _globals['_QUERY_SET_VALUES']._serialized_start=1752
  _globals['_QUERY_SET_VALUES']._serialized_end=2301
  _globals['_QUERY_SET_VALUES_INTS']._serialized_start=2118
  _globals['_QUERY_SET_VALUES_INTS']._serialized_end=2140
  _globals['_QUERY_SET_VALUES_DOUBLES']._serialized_start=2142
  _globals['_QUERY_SET_VALUES_DOUBLES']._serialized_end=2167
  _globals['_QUERY_SET_VALUES_STRINGS']._serialized_start=2169
  _globals['_QUERY_SET_VALUES_STRINGS']._serialized_end=2194
  _globals['_QUERY_SET_VALUES_BOOLS']._serialized_start=2196
  _globals['_QUERY_SET_VALUES_BOOLS']._serialized_end=2219
  _globals['_QUERY_SET_VALUES_LOCATIONS']._serialized_start=2221
  _globals['_QUERY_SET_VALUES_LOCATIONS']._serialized_end=2291
  _globals['_QUERY_SET_OPERATOR']._serialized_start=2303
  _globals['_QUERY_SET_OPERATOR']._serialized_end=2332
  _globals['_QUERY_CONSTRAINTEXPR']._serialized_start=2335
  _globals['_QUERY_CONSTRAINTEXPR']._serialized_end=3170
  _globals['_QUERY_CONSTRAINTEXPR_OR']._serialized_start=2640
  _globals['_QUERY_CONSTRAINTEXPR_OR']._serialized_end=2713
  _globals['_QUERY_CONSTRAINTEXPR_AND']._serialized_start=2715
  _globals['_QUERY_CONSTRAINTEXPR_AND']._serialized_end=2789
  _globals['_QUERY_CONSTRAINTEXPR_NOT']._serialized_start=2791
  _globals['_QUERY_CONSTRAINTEXPR_NOT']._serialized_end=2865
  _globals['_QUERY_CONSTRAINTEXPR_CONSTRAINT']._serialized_start=2868
  _globals['_QUERY_CONSTRAINTEXPR_CONSTRAINT']._serialized_end=3156
  _globals['_QUERY_MODEL']._serialized_start=3173
  _globals['_QUERY_MODEL']._serialized_end=3309
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Run different checks on AEA packages."""

import dis
import importlib
import os
import pprint
import re
import sys
from abc import abstractmethod
from collections import defaultdict
from functools import partial, wraps
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Set, Tuple, Union

import click
import yaml
from pip._internal.commands.show import search_packages_info

from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx
from aea.configurations.base import PackageId, PackageType, PublicId
from aea.configurations.constants import (
    AGENTS,
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_PROTOCOL_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
)


CONFIG_FILE_NAMES = [
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_SKILL_CONFIG_FILE,
    DEFAULT_CONNECTION_CONFIG_FILE,
    DEFAULT_CONTRACT_CONFIG_FILE,
    DEFAULT_PROTOCOL_CONFIG_FILE,
]  # type: List[str]

IGNORE_PACKAGES: Set[str] = {"pkg_resources"}
IGNORE_PACKAGE_SUBFOLDERS: Set[Path] = {Path("tests")}
DEP_NAME_RE = re.compile(r"(^[^=><\[]+)", re.I)  # type: ignore


class CustomException(Exception):
    """A custom exception class for this script."""

    @abstractmethod
    def print_error(self) -> None:
        """Print the error message."""


def list_decorator(fn: Callable) -> Callable:
    """Wraps generator to return list."""

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> List[Any]:
        return list(fn(*args, **kwargs))

    return wrapper


class PypiDependencyNotFound(CustomException):
    """Custom exception for Pypi dependency not found."""

    def print_error(self) -> None:
        """Print the error message."""
        return print(self.args[0])


class DependencyNotFound(CustomException):
    """Custom exception for dependencies not found."""

    def __init__(
        self,
        configuration_file: Path,
        expected_deps: Set[PackageId],
        missing_dependencies: Set[PackageId],
        *args: Any,
    ) -> None:
        """
        Initialize DependencyNotFound exception.

        :param configuration_file: path to the checked file.
        :param expected_deps: expected dependencies.
        :param missing_dependencies: missing dependencies.
        :param args: super class args.
        """
        super().__init__(*args)
        self.configuration_file = configuration_file
        self.expected_dependencies = expected_deps
        self.missing_dependencies = missing_dependencies

    def print_error(self) -> None:
        """Print the error message."""
        expected = list(map(str, self.expected_dependencies))
        missing = list(map(str, self.missing_dependencies))
        print("=" * 50)
        print(f"Package {self.configuration_file}:")
        print(f"Expected: {pprint.pformat(expected)}")
        print(f"Missing: {pprint.pformat(missing)}")
        print("=" * 50)


class EmptyPackageDescription(CustomException):
    """Custom exception for empty description field."""

    def __init__(
        self,
        configuration_file: Path,
        *args: Any,
    ) -> None:
        """
        Initialize EmptyPackageDescription exception.

        :param configuration_file: path to the checked file.
        :param args: super class args.
        """
        super().__init__(*args)
        self.configuration_file = configuration_file

    def print_error(self) -> None:
        """Print the error message."""
        print("=" * 50)
        print(f"Package '{self.configuration_file}' has empty description field.")
        print("=" * 50)


class UnexpectedAuthorError(CustomException):
    """Custom exception for unexpected author value."""

    def __init__(
        self,
        configuration_file: Path,
        expected_author: str,
        actual_author: str,
        *args: Any,
    ):
        """
        Initialize the exception.

        :param configuration_file: the file to the configuration that raised the error.
        :param expected_author: the expected author.
        :param actual_author: the actual author.
        :param args: other positional arguments.
        """
        super().__init__(*args)
        self.configuration_file = configuration_file
        self.expected_author = expected_author
        self.actual_author = actual_author

    def print_error(self) -> None:
        """Print the error message."""
        print("=" * 50)
        print(
            f"Package '{self.configuration_file}' has an unexpected author value: "
            f"expected {self.expected_author}, found '{self.actual_author}'."
        )
        print("=" * 50)


class PublicIdDefinitionError(CustomException):
    """Custom exception for error about PUBLIC_ID definitions in package Python modules."""

    def __init__(
        self,
        package_type: PackageType,
        public_id: PublicId,
        actual_nb_definitions: int,
        *args: Any,
    ) -> None:
        """Initialize the exception."""
        super().__init__(*args)
        self.package_type = package_type
        self.public_id = public_id
        self.actual_nb_definitions = actual_nb_definitions

    def print_error(self) -> None:
        """Print the error message."""
        print("=" * 50)
        print(
            f"expected unique definition of PUBLIC_ID for package {self.public_id} of type {self.package_type.value}; "
            f"found {self.actual_nb_definitions}"
        )
        print("=" * 50)


class WrongPublicIdError(CustomException):
    """Custom exception for error about wrong value of PUBLIC_ID."""

    def __init__(
        self,
        package_type: PackageType,
        public_id: PublicId,
        public_id_code: str,
        *args: Any,
    ) -> None:
        """Initialize the exception."""
        super().__init__(*args)
        self.package_type = package_type
        self.public_id = public_id
        self.public_id_code = public_id_code

    def print_error(self) -> None:
        """Print the error message."""
        print("=" * 50)
        print(
            f"expected {self.public_id} for package of type {self.package_type.value}; found '{self.public_id_code}'"
        )
        print("=" * 50)


def find_all_configuration_files(
    packages_dir: Path, vendor: Optional[str] = None
) -> List:
    """Find all configuration files."""
    config_files = [
        path
        for path in packages_dir.glob("*/*/*/*.yaml")
        if any([file in str(path) for file in CONFIG_FILE_NAMES])
        and (vendor is None or path.parent.parent.parent.name == vendor)
    ]
    return config_files


def get_public_id_from_yaml(configuration_file: Path) -> PublicId:
    """
    Get the public id from yaml.

    :param configuration_file: the path to the config yaml
    :return: public id
    """
    data = unified_yaml_load(configuration_file)
    author = data.get("author", None)
    if not author:
        raise KeyError(f"No author field in {str(configuration_file)}")
    # handle the case when it's a package or agent config file.
    try:
        name = data["name"] if "name" in data else data["agent_name"]
    except KeyError:
        click.echo(f"No name or agent_name field in {str(configuration_file)}")
        raise
    version = data.get("version", None)
    if not version:
        raise KeyError(f"No version field in {str(configuration_file)}")
    return PublicId(author, name, version)


def find_all_packages_ids(packages_dir: Path) -> Set[PackageId]:
    """Find all packages ids."""
    package_ids: Set[PackageId] = set()
    for configuration_file in find_all_configuration_files(packages_dir):
        package_type = PackageType(configuration_file.parts[-3][:-1])
        package_public_id = get_public_id_from_yaml(configuration_file)
        package_id = PackageId(package_type, package_public_id)
        package_ids.add(package_id)

    return package_ids


def unified_yaml_load(configuration_file: Path) -> Dict:
    """
    Load YAML file, unified (both single- and multi-paged).

    :param configuration_file: the configuration file path.
    :return: the data.
    """
    package_type = configuration_file.parent.parent.name
    with configuration_file.open() as fp:
        if package_type != AGENTS:
            return yaml.safe_load(fp)
        # when it is an agent configuration file,
        # we are interested only in the first page of the YAML,
        # because the dependencies are contained only there.
        data = yaml.safe_load_all(fp)
        return list(data)[0]


def check_dependencies(
    configuration_file: Path, all_packages_ids: Set[PackageId]
) -> None:
    """
    Check dependencies of configuration file.

    :param configuration_file: path to a package configuration file.
    :param all_packages_ids: all the package ids.
    """
    data = unified_yaml_load(configuration_file)

    def _add_package_type(package_type: PackageType, public_id_str: str) -> PackageId:
        return PackageId(package_type, PublicId.from_str(public_id_str))

    def _get_package_ids(
        package_type: PackageType, public_ids: Set[PublicId]
    ) -> Set[PackageId]:
        return set(map(partial(_add_package_type, package_type), public_ids))

    dependencies: Set[PackageId] = set.union(
        *[
            _get_package_ids(package_type, data.get(package_type.to_plural(), set()))
            for package_type in list(PackageType)
        ]
    )

    diff = dependencies.difference(all_packages_ids)
    if len(diff) > 0:
        raise DependencyNotFound(configuration_file, dependencies, diff)


def check_description(configuration_file: Path) -> None:
    """Check description field of a package is non-empty."""
    yaml_object = unified_yaml_load(configuration_file)
    description = yaml_object.get("description")
    if description == "":
        raise EmptyPackageDescription(configuration_file)


def check_author(configuration_file: Path, expected_author: str) -> None:
    """Check the author matches a certain desired value."""
    yaml_object = unified_yaml_load(configuration_file)
    actual_author = yaml_object.get("author", "")
    if actual_author != expected_author:
        raise UnexpectedAuthorError(configuration_file, expected_author, actual_author)


def check_public_id(configuration_file: Path) -> None:
    """Check the public_id in the code and configuration match."""
    expected_public_id = get_public_id_from_yaml(configuration_file)
    # remove last 's' character (as package type is plural in packages directory)
    package_type_str = configuration_file.parent.parent.name[:-1]
    package_type = PackageType(package_type_str)
    if package_type == PackageType.CONNECTION:
        module_name_to_load = Path("connection.py")
    elif package_type == PackageType.SKILL:
        module_name_to_load = Path("__init__.py")
    else:
        # no check to do.
        return
    module_path_to_load = configuration_file.parent / module_name_to_load
    content = module_path_to_load.read_text()

    # check number of definitions of PUBLIC_ID. Required exactly one match.
    assignments_to_public_id = re.findall("^PUBLIC_ID = (.*)", content, re.MULTILINE)
    if len(assignments_to_public_id) != 1:
        raise PublicIdDefinitionError(
            package_type, expected_public_id, len(assignments_to_public_id)
        )

    # check first pattern of public id: PublicId.from_str(...)
    matches = re.findall(
        r"^PUBLIC_ID = PublicId.from_str\( *(\"(.*)\"|'(.*)') *\)$",
        content,
        re.MULTILINE,
    )
    if len(matches) == 1:
        # process the result
        _, match1, match2 = matches[0]
        match = match1 if match1 != "" else match2
        if str(expected_public_id) != match:
            raise WrongPublicIdError(package_type, expected_public_id, match)
        return

    # check second pattern of public id: PublicId('...', '...', '...')
    matches = re.findall(
        r"^PUBLIC_ID = PublicId\( *['\"](.*)['\"] *, *['\"](.*)['\"] *, *['\"](.*)['\"] *\)$",
        content,
        re.MULTILINE,
    )
    if len(matches) == 1:
        # process the result
        author, name, version = matches[0]
        actual_public_id_str = f"{author}/{name}:{version}"
        if str(expected_public_id) != actual_public_id_str:
            raise WrongPublicIdError(
                package_type, expected_public_id, actual_public_id_str
            )
        return

    public_id_code = matches[0]
    if str(expected_public_id) not in public_id_code:
        raise WrongPublicIdError(package_type, expected_public_id, public_id_code)


class DependenciesTool:
    """Tool to work with setup.py dependencies."""

    @staticmethod
    def get_package_files(package_name: str) -> List[Path]:
        """Get package files list."""
        packages_info = list(search_packages_info([package_name]))
        if len(packages_info) == 0:
            raise Exception(f"package {package_name} not found")
        if isinstance(packages_info[0], dict):
            files = packages_info[0]["files"]
            location = packages_info[0]["location"]
        else:
            files = packages_info[0].files  # type: ignore
            location = packages_info[0].location  # type: ignore
        return [Path(location) / i for i in files]  # type: ignore

    @staticmethod
    def clean_dependency_name(dependecy_specification: str) -> str:
        """Get dependency name from dependency specification."""
        match = DEP_NAME_RE.match(dependecy_specification)
        if not match:
            raise ValueError(f"Bad dependency specification: {dependecy_specification}")
        return match.groups()[0]


class ImportsTool:
    """Tool to work with 3rd part imports in source code."""

    @staticmethod
    def get_imports_for_file(pyfile: Union[str, Path]) -> List[str]:
        """Get all imported modules for python source file."""
        with open(pyfile, "r") as f:
            statements = f.read()
        instructions = dis.get_instructions(statements)  # type: ignore
        imports = []
        is_try_except_block = False
        for im in instructions:
            if "SETUP_FINALLY" in im.opname:
                is_try_except_block = True

            if "POP_EXCEPT" in im.opname:
                is_try_except_block = False

            if is_try_except_block:
                continue

            if "IMPORT" in im.opname:
                imports.append(im)

        grouped: Dict[str, List[str]] = defaultdict(list)
        for instr in imports:
            grouped[instr.opname].append(instr.argval)

        return grouped["IMPORT_NAME"]

    @staticmethod
    def get_module_file(module_name: str) -> str:
        """Get module source file name."""
        try:
            mod = importlib.import_module(module_name)
            return getattr(mod, "__file__", "")
        except (AttributeError, ModuleNotFoundError):
            return ""

    @staticmethod
    @list_decorator
    def list_all_pyfiles(
        root_path: Union[Path, str],
        pattern: str = "**/*.py",
        ignores: Optional[Set[Path]] = None,
    ) -> Generator:
        """List all python files in directory."""
        ignores = ignores or set()
        root_path = Path(root_path)
        for path in root_path.glob(pattern):
            # check if path is a subpath of an ignore path
            relative_path = path.relative_to(root_path)
            if any(os.path.commonprefix([relative_path, p]) != "" for p in ignores):
                continue
            yield path.relative_to(root_path)

    @classmethod
    @list_decorator
    def get_third_part_imports_for_file(cls, pyfile: str) -> Generator:
        """Get list of third part modules imported for source file."""
        imports = cls.get_imports_for_file(pyfile)
        for module_name in imports:
            pyfile = cls.get_module_file(module_name)
            if not pyfile:
                continue
            if "site-packages" not in Path(pyfile).parts or "aea" in Path(pyfile).parts:
                continue
            yield module_name, Path(pyfile)

    @classmethod
    @list_decorator
    def list_all_pyfiles_with_third_party_imports(
        cls, root_path: Union[str, Path], pattern: str = "**/*.py"
    ) -> Generator:
        """Get list of all python sources with 3rd party modules imported."""
        for pyfile in cls.list_all_pyfiles(
            root_path, pattern=pattern, ignores=IGNORE_PACKAGE_SUBFOLDERS
        ):
            mods = list(cls.get_third_part_imports_for_file(root_path / pyfile))
            if not mods:
                continue
            yield Path(pyfile), list(set(mods))


class PyPIDependenciesCheckTool:
    """Tool to check imports in sources match dependencies in package configuration file."""

    def __init__(self, configuration_file: Path) -> None:
        """Init the checker for PyPI dependencies."""
        self.configuration_file = configuration_file
        self.config = unified_yaml_load(configuration_file)

    @staticmethod
    def make_third_party_imports(
        files_and_modules: List[Tuple[str, List[Tuple[str, Path]]]],
    ) -> Dict[str, Path]:
        """Make list of third party imports."""
        third_party_imports: Dict[str, Path] = {}
        for _pyfile, name_modules_pairs in files_and_modules:
            for package_name, modules in name_modules_pairs:
                third_party_imports[package_name] = modules

        return third_party_imports

    def get_dependencies(self) -> Dict[str, List[Path]]:
        """Get sections with dependencies with files lists."""
        dependencies = self.config.get("dependencies", set())
        result: Dict[str, List[Path]] = defaultdict(list)
        for dep in dependencies:
            dep = DependenciesTool.clean_dependency_name(dep)
            result[dep] = DependenciesTool.get_package_files(dep)
        return result

    def run(self) -> None:
        """Run dependency check."""
        files_and_modules = ImportsTool.list_all_pyfiles_with_third_party_imports(
            self.configuration_file.parent
        )
        package_dependencies = self.get_dependencies()
        third_party_imports = self.make_third_party_imports(files_and_modules)
        missed_deps_for_imports, deps_not_imported_directly = self.check_imports(
            third_party_imports, package_dependencies
        )

        if missed_deps_for_imports:
            raise PypiDependencyNotFound(
                f"unresolved imports: {', '.join(missed_deps_for_imports)}"
            )

        if deps_not_imported_directly:
            raise PypiDependencyNotFound(
                f"Dependencies not imported in code directly: {', '.join(deps_not_imported_directly)}"
            )

    @staticmethod
    def check_imports(
        third_party_imports: Dict[str, Path],
        pypi_dependencies: Dict[str, List[Path]],
    ) -> Tuple[Set[str], Set[str]]:
        """Find missing dependencies for imports and not imported dependencies."""

        def _find_dependency_for_module(
            dependencies: Dict[str, List[Path]], pyfile: Path
        ) -> Optional[str]:
            for package, files in dependencies.items():
                if pyfile in files:
                    return package
            return None

        imports_packages: Dict[str, Optional[str]] = {}
        for module, pyfile in third_party_imports.items():
            package_or_none = _find_dependency_for_module(pypi_dependencies, pyfile)
            if module not in IGNORE_PACKAGES:
                imports_packages[module] = package_or_none

        all_dependencies_set = set(third_party_imports.keys())
        used_dependencies_set = {
            k for k, v in imports_packages.items() if v is not None
        }
        deps_not_imported_directly: Set[str] = (
            all_dependencies_set - used_dependencies_set
        )
        missed_deps_for_imports: Set[str] = {
            k for k, v in imports_packages.items() if v is None
        }
        return missed_deps_for_imports, deps_not_imported_directly


def check_pypi_dependencies(configuration_file: Path) -> None:
    """Check PyPI dependencies of the AEA package."""
    PyPIDependenciesCheckTool(configuration_file).run()


@click.command(name="check-packages")
@click.option("--vendor", type=str, default=None, required=False)
@pass_ctx
def check_packages(ctx: Context, vendor: Optional[str]) -> None:
    """Run different checks on AEA packages."""
    packages_dir = Path(ctx.registry_path).absolute()
    all_packages_ids_ = find_all_packages_ids(packages_dir)
    failed: bool = False

    for file in find_all_configuration_files(packages_dir, vendor=vendor):
        try:
            expected_author = file.parent.parent.parent.name
            click.echo("Processing " + str(file))
            check_author(file, expected_author)
            check_dependencies(file, all_packages_ids_)
            check_description(file)
            check_public_id(file)
            check_pypi_dependencies(file)
        except CustomException as exception:
            exception.print_error()
            failed = True

    if failed:
        click.echo("Failed!")
        sys.exit(1)
    else:
        click.echo("OK!")
        sys.exit(0)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Memory usage of huge amount of messages."""
import time
from typing import Any, List, Tuple, Union

from aea_cli_benchmark.utils import get_mem_usage_in_mb  # noqa: I100

from aea.protocols.base import Message

from packages.fetchai.protocols.default.message import DefaultMessage


def make_message() -> Message:
    """Create a message."""
    return DefaultMessage(
        dialogue_reference=("", ""),
        message_id=1,
        target=0,
        performative=DefaultMessage.Performative.BYTES,
        content=b"",
    )


def run(messages_amount: int) -> List[Tuple[str, Union[int, float]]]:
    """Test messages generation and memory consumption."""
    messages: List[Any] = [
        0 for i in range(messages_amount)
    ]  # generate dummy list to count list structure memory
    mem_usage_on_start = get_mem_usage_in_mb()

    start_time = time.time()
    for i in range(messages_amount):
        messages[i] = make_message()
    mem_usage = get_mem_usage_in_mb()

    return [
        ("Mem usage(Mb)", mem_usage - mem_usage_on_start),
        ("Time (seconds)", time.time() - start_time),
    ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
Bump the AEA version throughout the code base.

usage: bump_aea_version [-h] [--new-version NEW_VERSION]
                        [-p KEY=VALUE [KEY=VALUE ...]] [--no-fingerprints]
                        [--only-check]

optional arguments:
  -h, --help            show this help message and exit
  --new-version NEW_VERSION
                        The new AEA version.
  -p KEY=VALUE [KEY=VALUE ...], --plugin-new-version KEY=VALUE [KEY=VALUE ...]
                        Set a number of key-value pairs plugin-name=new-
                        plugin-version
  --no-fingerprints     Skip the computation of fingerprints.
  --only-check          Only check the need of upgrade.


Example of usage:

python scripts/bump_aea_version.py --new-version 1.1.0 -p open-aea-ledger-fetchai=2.0.0 -p open-aea-ledger-ethereum=3.0.0
python scripts/bump_aea_version.py --only-check
"""

import argparse
import logging
import operator
import re
import sys
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, cast

from git import Repo  # pylint: disable=import-error
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from aea.cli.ipfs_hash import update_hashes
from aea.helpers.base import compute_specifier_from_version


logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(name)s][%(levelname)s] %(message)s"
)

# if the key is a file, just process it
# if the key is a directory, process all files below it
PatternByPath = Dict[Path, str]

AEA_DIR = Path("aea")
ROOT_DIR = Path(__file__).parent.parent

PLUGINS_DIR = Path("plugins")
ALL_PLUGINS = tuple(PLUGINS_DIR.iterdir())

"""
This pattern captures a specifier set in the dependencies section
of an AEA package configuration file, e.g.:

dependencies:
    ...
    open-aea-ledger-fetchai:
        version: >=1.0.0,<2.0.0
"""
YAML_DEPENDENCY_SPECIFIER_SET_PATTERN = (
    "(?<={package_name}:\n    version: )({specifier_set})"
)

"""
This pattern captures a specifier set for PyPI dependencies
in JSON format.

e.g.:
"open-aea-ledger-fetchai": {"version": ">=2.0.0, <3.0.0"}
"""
JSON_DEPENDENCY_SPECIFIER_SET_PATTERN = (
    '(?<="{package_name}": ."version": ")({specifier_set})(?=".)'
)


_AEA_ALL_PATTERN = r"(?<={package_name}\[all\]==){version}"
AEA_PATHS: PatternByPath = {
    Path("deploy-image", "Dockerfile"): _AEA_ALL_PATTERN,
    Path("deploy-image", "README.md"): "(?<=open-aea/tags/v){version}",
    Path("develop-image", "docker-env.sh"): "(?<=aea-develop:){version}",
    Path("docs", "quickstart.md"): "(?<=v){version}",
    Path("docs", "quickstart.md"): "(?<=open-aea/tags/v){version}",
    Path("examples", "tac_deploy", "Dockerfile"): _AEA_ALL_PATTERN,
    Path("scripts", "install.ps1"): _AEA_ALL_PATTERN,
    Path("scripts", "install.sh"): _AEA_ALL_PATTERN,
    Path(
        "tests", "test_docs", "test_bash_yaml", "md_files", "bash-quickstart.md"
    ): "(?<=v){version}",
    Path(
        "tests", "test_docs", "test_bash_yaml", "md_files", "bash-quickstart.md"
    ): "(?<=open-aea/tags/v){version}",
    Path("user-image", "docker-env.sh"): "(?<=aea-user:){version}",
    Path("skaffold.yaml"): "{version}",
}


def check_executed(func: Callable) -> Callable:
    """Check a functor has been already executed; if yes, raise error."""

    @wraps(func)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> None:
        if self.is_executed:
            raise ValueError("already executed")
        self._executed = True  # pylint: disable=protected-access
        self._result = func(self, *args, **kwargs)  # pylint: disable=protected-access

    return wrapper


def compute_specifier_from_version_custom(version: Version) -> str:
    """
    Post-process aea.helpers.compute_specifier_from_version

    The output is post-process in the following way:
    - remove spaces between specifier sets
    - put upper bound before lower bound

    :param version: the version
    :return: the specifier set according to the version and semantic versioning.
    """
    specifier_set_str = compute_specifier_from_version(version)
    specifiers = SpecifierSet(specifier_set_str)
    upper, lower = sorted(specifiers, key=str)
    return f"{upper},{lower}"


def get_regex_from_specifier_set(specifier_set: str) -> str:
    """
    Get the regex for specifier sets.

    This function accepts input of the form:

        ">={lower_bound_version}, <{upper_bound_version}"

    And computes a regex pattern:

        ">={lower_bound_version}, *<{upper_bound_version}|<{upper_bound_version}, *>={lower_bound_version}"

    i.e. not considering the order of the specifiers.

    :param specifier_set: The string representation of the specifier set
    :return: a regex pattern
    """
    specifiers = SpecifierSet(specifier_set)
    upper, lower = sorted(specifiers, key=str)
    alternatives = list()
    alternatives.append(f"{upper} *, *{lower}")
    alternatives.append(f"{lower} *, *{upper}")
    return "|".join(alternatives)


class PythonPackageVersionBumper:
    """Utility class to bump Python package versions."""

    IGNORE_DIRS = (Path(".git"),)

    def __init__(
        self,
        root_dir: Path,
        python_pkg_dir: Path,
        new_version: Version,
        files_to_pattern: PatternByPath,
        specifier_set_patterns: Sequence[str],
        package_name: Optional[str] = None,
        ignore_dirs: Sequence[Path] = (),
    ):
        """
        Initialize the utility class.

        :param root_dir: the root directory from which to look for files.
        :param python_pkg_dir: the path to the Python package to upgrade.
        :param new_version: the new version.
        :param files_to_pattern: a list of pairs.
        :param specifier_set_patterns: a list of patterns for specifier sets.
        :param package_name: the Python package name aliases (defaults to dirname of python_pkg_dir).
        :param ignore_dirs: a list of paths to ignore during the substitution.
        """
        self.root_dir = root_dir
        self.python_pkg_dir = python_pkg_dir
        self.new_version = new_version
        self.files_to_pattern = files_to_pattern
        self.specifier_set_patterns = specifier_set_patterns
        self.package_name = package_name or self.python_pkg_dir.name
        self.ignore_dirs = ignore_dirs or self.IGNORE_DIRS

        self.repo = Repo(self.root_dir)
        self._current_version: Optional[str] = None

        # functor pattern
        self._executed: bool = False
        self._result: Optional[bool] = None

    @property
    def is_executed(self) -> bool:
        """
        Return true if the functor has been executed; false otherwise.

        :return: True if it has been executed, False otherwise.
        """
        return self._executed

    @property
    def result(self) -> bool:
        """Get the result."""
        if not self.is_executed:
            raise ValueError("not executed yet")
        return cast(bool, self._result)

    @check_executed
    def run(self) -> bool:
        """Main entrypoint."""
        if not self.is_different_from_latest_tag():
            logging.info(
                f"The package {self.python_pkg_dir} has no changes since last tag."
            )
            return False
        new_version_string = str(self.new_version)
        current_version_str = self.update_version_for_package(new_version_string)

        # validate current version
        current_version: Version = Version(current_version_str)
        current_version_str = str(current_version)
        self._current_version = current_version_str
        self.update_version_for_files()

        return self.update_version_specifiers(current_version, self.new_version)

    def update_version_for_files(self) -> None:
        """Update the version."""
        for filepath, regex_template in self.files_to_pattern.items():
            self.update_version_for_file(
                filepath,
                cast(str, self._current_version),
                str(self.new_version),
                version_regex_template=regex_template,
            )

    def update_version_for_package(self, new_version: str) -> str:
        """
        Update version for file.

        If __version__.py is available, parse it and check for __version__ variable.
        Otherwise, try to parse setup.py.
        Otherwise, raise error.

        :param new_version: the new version
        :return: the current version
        """
        version_path = self.python_pkg_dir / Path("__version__.py")
        setup_path = self.python_pkg_dir.parent / "setup.py"
        if version_path.exists():
            regex_template = '(?<=__version__ = [\'"])({version})(?=")'
            path = version_path
        elif setup_path.exists():
            regex_template = r'(?<=version=[\'"])({version})(?=[\'"],)'
            path = setup_path
        else:
            raise ValueError("cannot fine neither '__version__.py' nor 'setup.py'")

        content = path.read_text()
        pattern = regex_template.format(version=".*")
        current_version_candidates = re.findall(pattern, content)
        more_than_one_match = len(current_version_candidates) > 1
        if more_than_one_match:
            raise ValueError(
                f"find more than one match for current version in {path}: {current_version_candidates}"
            )
        current_version = current_version_candidates[0]
        self.update_version_for_file(
            path, current_version, new_version, version_regex_template=regex_template
        )
        return current_version

    def update_version_for_file(
        self,
        path: Path,
        current_version: str,
        new_version: str,
        version_regex_template: Optional[str] = None,
    ) -> None:
        """
        Update version for file.

        :param path: the file path
        :param current_version: the regex for the current version
        :param new_version: the new version
        :param version_regex_template: the regex template to replace with the current version. Defaults to exactly the current version.
        """
        if version_regex_template is not None:
            regex_str = version_regex_template.format(
                package_name=self.package_name, version=current_version
            )
        else:
            regex_str = current_version
        pattern = re.compile(regex_str)
        content = path.read_text()
        content = pattern.sub(new_version, content)
        path.write_text(content)

    def update_version_specifiers(
        self, old_version: Version, new_version: Version
    ) -> bool:
        """
        Update specifier set.

        :param old_version: the old version.
        :param new_version: the new version.
        :return: True if the update has been done, False otherwise.
        """
        old_specifier_set = compute_specifier_from_version_custom(old_version)
        new_specifier_set = compute_specifier_from_version_custom(new_version)
        logging.info(f"Old version specifier: {old_specifier_set}")
        logging.info(f"New version specifier: {new_specifier_set}")
        if old_specifier_set == new_specifier_set:
            logging.info("Not updating version specifier - they haven't changed.")
            return False
        for file in filter(lambda p: not p.is_dir(), self.root_dir.rglob("*")):
            dir_root = Path(file.parts[0])
            if dir_root in self.ignore_dirs:
                logging.info(f"Skipping '{file}'...")
                continue
            logging.info(
                f"Replacing '{old_specifier_set}' with '{new_specifier_set}' in '{file}'... ",
            )
            try:
                content = file.read_text()
            except UnicodeDecodeError as e:
                logging.info(f"Cannot read {file}: {str(e)}. Continue...")
            else:
                content = self._replace_specifier_sets(
                    old_specifier_set, new_specifier_set, content
                )
                file.write_text(content)
        return True

    def _replace_specifier_sets(
        self, old_specifier_set: str, new_specifier_set: str, content: str
    ) -> str:
        old_specifier_set_regex = get_regex_from_specifier_set(old_specifier_set)
        for pattern_template in self.specifier_set_patterns:
            regex = pattern_template.format(
                package_name=self.package_name,
                specifier_set=old_specifier_set_regex,
            )
            pattern = re.compile(regex)
            if pattern.search(content) is not None:
                content = pattern.sub(new_specifier_set, content)
        return content

    def is_different_from_latest_tag(self) -> bool:
        """Check whether the package has changes since the latest tag."""
        assert len(self.repo.tags) > 0, "no git tags found"
        latest_tag_str = str(self.repo.tags[-1])
        args = latest_tag_str, "--", str(self.python_pkg_dir)
        logging.info(f"Running 'git diff {' '.join(args)}'")
        diff = self.repo.git.diff(*args)
        return diff != ""


def parse_args() -> argparse.Namespace:
    """Parse arguments."""

    parser = argparse.ArgumentParser("bump_aea_version")
    parser.add_argument(
        "--new-version", type=str, required=False, help="The new AEA version."
    )
    parser.add_argument(
        "-p",
        "--plugin-new-version",
        metavar="KEY=VALUE",
        nargs="+",
        help="Set a number of key-value pairs plugin-name=new-plugin-version",
        default={},
    )
    parser.add_argument(
        "--no-fingerprints",
        action="store_true",
        help="Skip the computation of fingerprints.",
    )
    parser.add_argument(
        "--only-check", action="store_true", help="Only check the need of upgrade."
    )
    arguments_ = parser.parse_args()
    return arguments_


def make_aea_bumper(new_aea_version: Version) -> PythonPackageVersionBumper:
    """Build the AEA Python package version bumper."""
    aea_version_bumper = PythonPackageVersionBumper(
        ROOT_DIR,
        AEA_DIR,
        new_aea_version,
        specifier_set_patterns=[
            "(?<=aea_version:) *({specifier_set})",
            "(?<={package_name})({specifier_set})",
        ],
        files_to_pattern=AEA_PATHS,
    )
    return aea_version_bumper


def make_plugin_bumper(
    plugin_dir: Path, new_version: Version
) -> PythonPackageVersionBumper:
    """Build the plugin Python package version bumper."""
    plugin_package_dir = plugin_dir / plugin_dir.name.replace("-", "_")
    plugin_version_bumper = PythonPackageVersionBumper(
        ROOT_DIR,
        plugin_package_dir,
        new_version,
        files_to_pattern={},
        specifier_set_patterns=[
            YAML_DEPENDENCY_SPECIFIER_SET_PATTERN,
            JSON_DEPENDENCY_SPECIFIER_SET_PATTERN,
        ],
        package_name=plugin_dir.name,
    )
    return plugin_version_bumper


def process_plugins(new_versions: Dict[str, Version]) -> bool:
    """Process plugins."""
    result = False
    for plugin_dir in ALL_PLUGINS:
        plugin_dir_name = plugin_dir.name
        if plugin_dir_name not in new_versions:
            logging.info(
                f"Skipping {plugin_dir_name} as it is not specified in input {new_versions}"
            )
            continue
        new_version = new_versions[plugin_dir_name]
        logging.info(
            f"Processing {plugin_dir_name}: upgrading at version {new_version}"
        )
        plugin_bumper = make_plugin_bumper(plugin_dir, new_version)
        plugin_bumper.run()
        result |= plugin_bumper.result
    return result


def parse_plugin_versions(key_value_strings: List[str]) -> Dict[str, Version]:
    """Parse plugin versions."""
    return {
        plugin_name: Version(version)
        for plugin_name, version in map(
            operator.methodcaller("split", "="), key_value_strings
        )
    }


def only_check_bump_needed() -> int:
    """
    Check whether a version bump is needed for AEA and plugins.

    :return: the return code
    """
    bumpers: List[PythonPackageVersionBumper] = list()
    to_upgrade: List[Path] = list()
    bumpers.append(make_aea_bumper(None))  # type: ignore
    for plugin_dir in ALL_PLUGINS:
        bumpers.append(make_plugin_bumper(plugin_dir, None))  # type: ignore

    latest_tag = str(bumpers[0].repo.tags[-1])
    logging.info(
        f"Checking packages that have changes from tag {latest_tag} and that require a new release..."
    )
    for bumper in bumpers:
        if bumper.is_different_from_latest_tag():
            logging.info(
                f"Package {bumper.python_pkg_dir} is different from latest tag {latest_tag}."
            )
            to_upgrade.append(bumper.python_pkg_dir)

    if len(to_upgrade) > 0:
        logging.info("Packages to upgrade:")
        for path in to_upgrade:
            logging.info(path)
    else:
        logging.info("No packages to upgrade.")
    return 0


def bump(arguments: argparse.Namespace) -> int:
    """
    Bump versions.

    :param arguments: arguments from argparse
    :return: the return code
    """
    new_plugin_versions = parse_plugin_versions(arguments.plugin_new_version)
    logging.info(f"Parsed arguments: {arguments}")
    logging.info(f"Parsed plugin versions: {new_plugin_versions}")

    have_updated_specifier_set = False
    if arguments.new_version is not None:
        new_aea_version = Version(arguments.new_version)
        aea_version_bumper = make_aea_bumper(new_aea_version)
        aea_version_bumper.run()
        have_updated_specifier_set = aea_version_bumper.result
        logging.info("AEA package processed.")
    else:
        logging.info("AEA package not processed - no version provided.")

    logging.info("Processing plugins:")
    have_updated_specifier_set |= process_plugins(new_plugin_versions)

    logging.info("OK")
    return_code = 0
    if arguments.no_fingerprints:
        logging.info(
            "Not updating fingerprints, since --no-fingerprints was specified."
        )
    elif have_updated_specifier_set is False:
        logging.info(
            "Not updating fingerprints, since no specifier set has been updated."
        )
    else:
        logging.info("Updating hashes and fingerprints.")
        return_code = update_hashes(
            packages_dir=ROOT_DIR / "packages",
        )
    return return_code


def main() -> None:
    """Run the script."""
    repo = Repo(str(ROOT_DIR))
    if repo.is_dirty():
        logging.info(
            "Repository is dirty. Please clean it up before running this script."
        )
        sys.exit(1)

    arguments = parse_args()
    if arguments.only_check:
        sys.exit(only_check_bump_needed())

    return_code = bump(arguments)
    sys.exit(return_code)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Envelopes generation speed for Behaviour act test."""
import itertools
import struct
import time
from typing import List, Tuple, Union, cast

from aea_cli_benchmark.utils import get_mem_usage_in_mb  # noqa: I100
from aea_cli_benchmark.utils import (
    make_agent,
    make_envelope,
    make_skill,
    wait_for_condition,
)

from aea.configurations.base import ConnectionConfig
from aea.identity.base import Identity
from aea.protocols.base import Message
from aea.registries.resources import Resources
from aea.runner import AEARunner
from aea.skills.base import Handler

from packages.fetchai.connections.local.connection import (  # noqa: E402 # pylint: disable=C0413
    LocalNode,
    OEFLocalConnection,
)
from packages.fetchai.protocols.default.message import DefaultMessage


class TestHandler(Handler):
    """Dummy handler to handle messages."""

    SUPPORTED_PROTOCOL = DefaultMessage.protocol_id

    def setup(self) -> None:
        """Noop setup."""
        self.count: int = 0  # pylint: disable=attribute-defined-outside-init
        self.rtt_total_time: float = (
            0.0  # pylint: disable=attribute-defined-outside-init
        )
        self.rtt_count: int = 0  # pylint: disable=attribute-defined-outside-init

        self.latency_total_time: float = (
            0.0  # pylint: disable=attribute-defined-outside-init
        )
        self.latency_count: int = 0  # pylint: disable=attribute-defined-outside-init

    def teardown(self) -> None:
        """Noop teardown."""

    def handle(self, message: Message) -> None:
        """Handle incoming message."""
        self.count += 1

        if message.dialogue_reference[0] != "":
            rtt_ts, latency_ts = struct.unpack("dd", message.content)  # type: ignore
            if message.dialogue_reference[0] == self.context.agent_address:
                self.rtt_total_time += time.time() - rtt_ts
                self.rtt_count += 1

            self.latency_total_time += time.time() - latency_ts
            self.latency_count += 1

        if message.dialogue_reference[0] in ["", self.context.agent_address]:
            # create new
            response_msg = DefaultMessage(
                dialogue_reference=(self.context.agent_address, ""),
                message_id=1,
                target=0,
                performative=DefaultMessage.Performative.BYTES,
                content=struct.pack("dd", time.time(), time.time()),
            )
        else:
            # update ttfb copy rtt
            response_msg = DefaultMessage(
                dialogue_reference=message.dialogue_reference,
                message_id=1,
                target=0,
                performative=DefaultMessage.Performative.BYTES,
                content=struct.pack("dd", rtt_ts, time.time()),  # type: ignore
            )

        self.context.outbox.put(make_envelope(message.to, message.sender, response_msg))


def run(
    duration: int,
    runtime_mode: str,
    runner_mode: str,
    start_messages: int,
    num_of_agents: int,
) -> List[Tuple[str, Union[int, float]]]:
    """Test multiagent message exchange."""
    # pylint: disable=import-outside-toplevel,unused-import
    # import manually due to some lazy imports in decision_maker
    import aea.decision_maker.default  # noqa: F401

    local_node = LocalNode()
    local_node.start()

    agents = []
    skills = []

    for i in range(num_of_agents):
        resources = Resources()
        agent_name = f"agent{i}"
        identity = Identity(agent_name, address=agent_name, public_key="somekey")
        connection = OEFLocalConnection(
            local_node,
            configuration=ConnectionConfig(
                connection_id=OEFLocalConnection.connection_id,
            ),
            identity=identity,
            data_dir="tmp",
        )
        resources.add_connection(connection)
        agent = make_agent(
            agent_name=agent_name,
            runtime_mode=runtime_mode,
            resources=resources,
            identity=identity,
        )
        skill = make_skill(agent, handlers={"test": TestHandler})
        agent.resources.add_skill(skill)
        agents.append(agent)
        skills.append(skill)

    runner = AEARunner(agents, runner_mode)
    runner.start(threaded=True)

    for agent_ in agents:
        wait_for_condition(lambda a=agent_: a.is_running, timeout=5)
    wait_for_condition(lambda: runner.is_running, timeout=5)
    time.sleep(1)

    for agent1, agent2 in itertools.permutations(agents, 2):
        env = make_envelope(agent1.identity.address, agent2.identity.address)

        for _ in range(int(start_messages)):
            agent1.outbox.put(env)

    time.sleep(duration)

    mem_usage = get_mem_usage_in_mb()

    local_node.stop()
    runner.stop(timeout=5)

    total_messages = sum(
        [cast(TestHandler, skill.handlers["test"]).count for skill in skills]
    )
    rate = total_messages / duration

    rtt_total_time = sum(
        [cast(TestHandler, skill.handlers["test"]).rtt_total_time for skill in skills]
    )
    rtt_count = sum(
        [cast(TestHandler, skill.handlers["test"]).rtt_count for skill in skills]
    )

    if rtt_count == 0:
        rtt_count = -1

    latency_total_time = sum(
        [
            cast(TestHandler, skill.handlers["test"]).latency_total_time
            for skill in skills
        ]
    )
    latency_count = sum(
        [cast(TestHandler, skill.handlers["test"]).latency_count for skill in skills]
    )

    if latency_count == 0:
        latency_count = -1

    return [
        ("Total Messages handled", total_messages),
        ("Messages rate(envelopes/second)", rate),
        ("Mem usage(Mb)", mem_usage),
        ("RTT (ms)", rtt_total_time / rtt_count),
        ("Latency (ms)", latency_total_time / latency_count),
    ]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

# pylint: skip-file

"""Conftest module for Pytest."""

import functools
import logging
import time
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Generator

import docker
import pytest
import yaml
from docker.errors import ImageNotFound, NotFound

from aea.test_tools.docker_image import Container, DockerImage

from packages.valory.connections import p2p_libp2p_client
from packages.valory.connections.test_libp2p.tests.acn_image import (
    ACNNodeDockerImage,
    ACNWithBootstrappedEntryNodesDockerImage,
)


@dataclass
class NodeConfig:
    """Node configuration"""

    uri: str
    maddr: str
    public_key: str


DOCKER_PRINT_SEPARATOR = ("\n" + "*" * 40) * 3 + "\n"

logger = logging.getLogger(__name__)


META_ADDRESS = "0.0.0.0"  # nosec
ACN_CONFIGURATION: Dict[str, str] = dict(
    AEA_P2P_ID="54562eb807d2f80df8151db0a394cac72e16435a5f64275c277cae70308e8b24",
    AEA_P2P_URI_PUBLIC=f"{META_ADDRESS}:5000",
    AEA_P2P_URI=f"{META_ADDRESS}:5000",
    AEA_P2P_DELEGATE_URI=f"{META_ADDRESS}:11000",
    AEA_P2P_URI_MONITORING=f"{META_ADDRESS}:8080",
    ACN_LOG_FILE="/acn/libp2p_node.log",
)

local_nodes = [
    NodeConfig(
        "localhost:11001",
        f"/dns4/{META_ADDRESS}/tcp/9001/p2p/16Uiu2HAkw99FW2GKb2qs24eLgfXSSUjke1teDaV9km63Fv3UGdnF",
        "02197b55d736bd242311aaabb485f9db40881349873bb13e8b60c8a130ecb341d8",
    ),
    NodeConfig(
        "localhost:11002",
        f"/dns4/{META_ADDRESS}/tcp/9002/p2p/16Uiu2HAm4aHr1iKR323tca8Zu8hKStEEVwGkE2gtCJw49S3gbuVj",
        "0287ee61e8f939aeaa69bd7156463d698f8e74a3e1d5dd20cce997970f13ad4f12",
    ),
]


@functools.lru_cache()
def load_client_connection_yaml_config() -> Dict[str, Any]:
    """Load libp2p client connection yaml configuration"""

    connection_yaml = (
        Path(p2p_libp2p_client.__file__).absolute().parent / "connection.yaml"
    )
    config = yaml.safe_load(connection_yaml.read_text())["config"]
    dns_template = "/dns4/{domain}/tcp/{port}/p2p/{peer_id}"

    # add matching information needed in ACN libp2p tests
    port_mapping = {
        "9005": ("9003", "16Uiu2HAm9ftkcmsBwPf2KXjrUd9G6GPi2WN3opXvjrjukNpE9e5k"),
        "9006": ("9004", "16Uiu2HAmAzQL3YV2Yv37ffafuMaaPtLUSPBEoyjDyFyrcLQzgB6P"),
    }

    for node in config["nodes"]:
        domain, port = node["uri"].split(":")
        port, peer_id = port_mapping[port]
        node["maddr"] = dns_template.format(domain=domain, port=port, peer_id=peer_id)

    return config


LIBP2P_LEDGER = load_client_connection_yaml_config()["ledger_id"]
public_nodes = [
    NodeConfig(**kw) for kw in load_client_connection_yaml_config()["nodes"]
]


def _launch_image(
    image: DockerImage, timeout: float = 2.0, max_attempts: int = 10
) -> Generator:
    """Launch image."""

    image.check_skip()
    image.stop_if_already_running()
    container = image.create()
    container.start()
    logger.info(f"Setting up image {image.tag}...")
    success = image.wait(max_attempts, timeout)
    if not success:
        container.stop()
        container.remove()
        pytest.fail(f"{image.tag} doesn't work. Exiting...")
    else:
        try:
            logger.info("Done!")
            time.sleep(timeout)
            yield
        finally:
            logger.info(f"Stopping the image {image.tag}...")
            container.stop()
            container.remove()


def _pre_launch(image: DockerImage) -> None:
    """Run pre-launch checks."""
    image.check_skip()
    image.stop_if_already_running()


def _start_container(
    image: DockerImage, container, timeout: float, max_attempts: int
) -> None:
    """
    Start a container.

    :param image: an instance of Docker image.
    :param container: the container to start, created from the image.
    :param timeout: timeout to launch
    :param max_attempts: max launch attempts
    """
    container.start()
    logger.info(f"Setting up image {image.tag}...")
    success = image.wait(max_attempts, timeout)
    if not success:
        container.stop()
        logger.error(
            f"{DOCKER_PRINT_SEPARATOR}Logs from container {container.name}:\n{container.logs().decode()}"
        )
        container.remove()
        pytest.fail(f"{image.tag} doesn't work. Exiting...")
    else:
        logger.info("Done!")
        time.sleep(timeout)


def _stop_container(container: Container, tag: str) -> None:
    """Stop a container."""
    logger.info(f"Stopping container {container.name} from image {tag}...")
    container.stop()
    try:
        logger.info(
            f"{DOCKER_PRINT_SEPARATOR}Logs from container {container.name}:\n{container.logs().decode()}"
        )
        if str(container.name).startswith("node"):
            logger.info(
                f"{DOCKER_PRINT_SEPARATOR}Logs from container log file {container.name}:\n"
            )
            bits, _ = container.get_archive(f"/logs/{container.name}.txt")
            for chunk in bits:
                logger.info(chunk.decode())
    except (ImageNotFound, NotFound) as e:
        logger.error(e)
    finally:
        container.remove()


def launch_many_containers(
    image: DockerImage, timeout: float = 2.0, max_attempts: int = 10
) -> Generator[DockerImage, None, None]:
    """Launch many containers from an image."""
    _pre_launch(image)
    containers = image.create()
    for container in containers:
        _start_container(image, container, timeout, max_attempts)
    yield image
    for container in containers:
        _stop_container(container, image.tag)


@pytest.fixture(scope="session")
def acn_configuration():
    """Get the ACN Node configuration for testing purposes."""
    return ACN_CONFIGURATION


@contextmanager
def _acn_context(
    acn_configuration: Dict,
    timeout: float = 2.0,
    max_attempts: int = 10,
):
    client = docker.from_env()
    image = ACNNodeDockerImage(client, config=acn_configuration)
    yield from _launch_image(image, timeout=timeout, max_attempts=max_attempts)


@pytest.mark.integration
@pytest.mark.ledger
@pytest.fixture(scope="class")
def acn_node(
    acn_configuration,
    timeout: float = 2.0,
    max_attempts: int = 10,
):
    """Launch the ACN image."""
    with _acn_context(acn_configuration, timeout, max_attempts) as image:
        yield image


@pytest.mark.integration
class UseACNNode:
    """Inherit from this class to an ACN Node."""

    @pytest.fixture(autouse=True)
    def _start_acn_node(self, acn_node):
        """Start a ACN Node image."""


@contextmanager
def _acn_multiple_nodes_context(
    acn_configuration: Dict,
    timeout: float = 2.0,
    max_attempts: int = 10,
):
    client = docker.from_env()
    image = ACNWithBootstrappedEntryNodesDockerImage(client, config=acn_configuration)
    yield from launch_many_containers(image, timeout, max_attempts)


@pytest.fixture(scope="class")
def acn_multiple_nodes(
    acn_configuration,
    timeout: float = 2.0,
    max_attempts: int = 10,
):
    """Launch the ACN images."""
    with _acn_multiple_nodes_context(acn_configuration, timeout, max_attempts) as image:
        logging.info("Starting up local ACN nodes")
        yield image


@pytest.mark.integration
class ACNWithBootstrappedEntryNodes:
    """Inherit from this class to an ACN Node."""

    @pytest.fixture(autouse=True)
    def _start_acn(self, acn_multiple_nodes):
        """Start a series of ACN Node images."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the libp2p client connection."""
import asyncio
import hashlib
import logging
import random
import re
import ssl
from asyncio import CancelledError
from asyncio.streams import StreamWriter
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast
from urllib.parse import urlparse

import aiohttp
from aiohttp.client_reqrep import ClientResponse
from asn1crypto import x509  # type: ignore
from ecdsa.curves import SECP256k1
from ecdsa.keys import VerifyingKey
from ecdsa.util import sigdecode_der

from aea.configurations.base import PublicId
from aea.configurations.constants import DEFAULT_LEDGER
from aea.connections.base import Connection, ConnectionStates
from aea.crypto.registries import make_crypto
from aea.exceptions import enforce
from aea.helpers.acn.agent_record import AgentRecord
from aea.helpers.acn.uri import Uri
from aea.mail.base import Envelope

from packages.valory.protocols.acn import acn_pb2
from packages.valory.protocols.acn.message import AcnMessage


try:
    from asyncio.streams import (  # type: ignore  # pylint: disable=ungrouped-imports
        IncompleteReadError,
    )
except ImportError:  # pragma: nocover
    from asyncio import IncompleteReadError  # pylint: disable=ungrouped-imports


_default_logger = logging.getLogger("aea.packages.valory.connections.p2p_libp2p_client")

PUBLIC_ID = PublicId.from_str("valory/p2p_libp2p_mailbox:0.1.0")

SUPPORTED_LEDGER_IDS = ["fetchai", "cosmos", "ethereum"]

POR_DEFAULT_SERVICE_ID = "acn"

ACN_CURRENT_VERSION = "0.1.0"


class NodeClient:
    """Client to communicate with node using ipc channel(pipe)."""

    NO_ENVELOPES_SLEEP_TIME: float = 2.0

    def __init__(self, node_uri: Uri, node_por: AgentRecord) -> None:
        """Set node client with pipe."""
        self.node_uri = node_uri
        self.agent_record = node_por
        self._session_token: Optional[str] = None
        self.ssl_ctx = Optional[ssl.SSLContext]

    async def connect(self) -> bool:
        """Connect to node with pipe."""
        url = f"https://{self.node_uri}/ssl_signature"
        self.ssl_ctx = await SSLValidator(
            url, self.agent_record.representative_public_key
        ).check()
        await self.register()
        return True

    async def send_envelope(self, envelope: Envelope) -> None:
        """Send envelope to node."""
        if not self._session_token:  # pragma: nocover
            raise ValueError("not connected!")

        envelope_data = envelope.encode()
        # send here
        response, _ = await self._perform_http_request(
            method="POST",
            url="/send_envelope",
            data=envelope_data,
            headers={"Session-Id": self._session_token},
        )

        if response.status != 200:  # pragma: nocover
            text = await response.text()
            raise ValueError(f"Bad response code: {response.status} {text}")

    async def _perform_http_request(
        self, method: str, url: str, **kwargs: Any
    ) -> Tuple[ClientResponse, bytes]:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                method, url=f"https://{self.node_uri}{url}", ssl=self.ssl_ctx, **kwargs
            ) as response:
                data = await response.read()
                return response, data

    def make_agent_record(self) -> AcnMessage.AgentRecord:  # type: ignore
        """Make acn agent record."""
        agent_record = AcnMessage.AgentRecord(
            address=self.agent_record.address,
            public_key=self.agent_record.public_key,
            peer_public_key=self.agent_record.representative_public_key,
            signature=self.agent_record.signature,
            service_id=POR_DEFAULT_SERVICE_ID,
            ledger_id=self.agent_record.ledger_id,
        )
        return agent_record

    async def read_envelope(self) -> Optional[Envelope]:
        """Read envelope from the mailbox node."""
        while True:
            if not self._session_token:  # pragma: nocover
                raise ValueError("Client not registered!")
            response, data = await self._perform_http_request(
                "GET", "/get_envelope", headers={"Session-Id": self._session_token}
            )
            if response.status != 200:  # pragma: nocover
                raise ValueError(f"Bad response code: {response.status}")

            if not data:
                await asyncio.sleep(self.NO_ENVELOPES_SLEEP_TIME)
                continue

            envelope = Envelope.decode(data)
            return envelope

    async def register(self) -> None:
        """Register agent on the remote node."""
        agent_record = self.make_agent_record()
        performative = acn_pb2.AcnMessage.Register_Performative()  # type: ignore  # pylint: disable=no-member
        AcnMessage.AgentRecord.encode(
            performative.record, agent_record  # pylint: disable=no-member
        )
        data = performative.record.SerializeToString()  # pylint: disable=no-member
        response, _ = await self._perform_http_request(
            "POST", url="/register", data=data
        )

        if response.status != 200:  # pragma: nocover
            raise ValueError(f"Bad response code: {response.status}")

        token = await response.text()
        if not re.match("[0-9a-f]{32}", token, re.I):  # pragma: nocover
            raise ValueError(f"invalid response: {token}")

        self._session_token = token

    async def close(self) -> None:
        """Close node connection."""
        if not self._session_token:  # pragma: nocover
            raise ValueError("not connected!")

        response, _ = await self._perform_http_request(
            "GET", "/unregister", headers={"Session-Id": self._session_token}
        )
        if response.status != 200:  # pragma: nocover
            raise ValueError(f"Bad response code: {response.status}")

        self._session_token = None


class P2PLibp2pMailboxConnection(Connection):
    """
    A libp2p client connection.

    Send and receive envelopes to and from agents on the p2p network without deploying a libp2p node.
    Connect to the libp2p node using traffic delegation service.
    """

    connection_id = PUBLIC_ID

    DEFAULT_CONNECT_RETRIES = 3
    DEFAULT_TLS_CONNECTION_SIGNATURE_TIMEOUT = 5.0

    def __init__(self, **kwargs: Any) -> None:
        """Initialize a libp2p client connection."""
        super().__init__(**kwargs)

        self.tls_connection_signature_timeout = self.configuration.config.get(
            "tls_connection_signature_timeout",
            self.DEFAULT_TLS_CONNECTION_SIGNATURE_TIMEOUT,
        )
        self.connect_retries = self.configuration.config.get(
            "connect_retries", self.DEFAULT_CONNECT_RETRIES
        )
        ledger_id = self.configuration.config.get("ledger_id", DEFAULT_LEDGER)
        if ledger_id not in SUPPORTED_LEDGER_IDS:
            raise ValueError(  # pragma: nocover
                "Ledger id '{}' is not supported. Supported ids: '{}'".format(
                    ledger_id, SUPPORTED_LEDGER_IDS
                )
            )

        key_file: Optional[str] = self.configuration.config.get("tcp_key_file")
        nodes: Optional[List[Dict[str, Any]]] = self.configuration.config.get("nodes")

        if nodes is None:
            raise ValueError("At least one node should be provided")
        nodes = list(nodes)

        nodes_uris = [node.get("uri", None) for node in nodes]
        enforce(
            len(nodes_uris) == len(nodes) and None not in nodes_uris,
            "Delegate 'uri' should be provided for each node",
        )

        nodes_public_keys = [node.get("public_key", None) for node in nodes]
        enforce(
            len(nodes_public_keys) == len(nodes) and None not in nodes_public_keys,
            "Delegate 'public_key' should be provided for each node",
        )

        cert_requests = self.configuration.cert_requests
        if cert_requests is None or len(cert_requests) != len(nodes):
            raise ValueError(  # pragma: nocover
                "cert_requests field must be set and contain exactly as many entries as 'nodes'!"
            )
        for cert_request in cert_requests:
            save_path = cert_request.get_absolute_save_path(Path(self.data_dir))
            if not save_path.is_file():
                raise Exception(  # pragma: nocover
                    "cert_request 'save_path' field is not a file. "
                    "Please ensure that 'issue-certificates' command is called beforehand"
                )

        # we cannot use the key from the connection's crypto store as
        # the key will be used for TLS tcp connection, whereas the
        # connection's crypto store key is used for PoR
        if key_file is not None:
            key = make_crypto(ledger_id, private_key_path=key_file)
        else:
            key = make_crypto(ledger_id)

        # client connection id
        self.key = key
        self.logger.debug("Public key used for TCP: {}".format(key.public_key))

        # delegate uris
        self.delegate_uris = [Uri(node_uri) for node_uri in nodes_uris]

        # delegates PoRs
        self.delegate_pors: List[AgentRecord] = []
        for i, cert_request in enumerate(cert_requests):
            agent_record = AgentRecord.from_cert_request(
                cert_request, self.address, nodes_public_keys[i], self.data_dir
            )
            self.delegate_pors.append(agent_record)

        # select a delegate
        index = random.randint(0, len(self.delegate_uris) - 1)  # nosec
        self.node_uri = self.delegate_uris[index]
        self.node_por = self.delegate_pors[index]
        self.logger.debug("Node to use as delegate: {}".format(self.node_uri))

        self._in_queue = None  # type: Optional[asyncio.Queue]
        self._process_messages_task = None  # type: Optional[asyncio.Future]
        self._node_client: Optional[NodeClient] = None

        self._send_queue: Optional[asyncio.Queue] = None
        self._send_task: Optional[asyncio.Task] = None

    async def _send_loop(self) -> None:
        """Handle message in  the send queue."""

        if not self._send_queue or not self._node_client:  # pragma: nocover
            self.logger.error("Send loop not started cause not connected properly.")
            return
        try:
            while self.is_connected:
                envelope = await self._send_queue.get()
                await self._send_envelope_with_node_client(envelope)
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise  # pragma: nocover
        except Exception:  # pylint: disable=broad-except # pragma: nocover
            self.logger.exception(
                f"Failed to send an envelope {envelope}. Stop connection."
            )
            await asyncio.shield(self.disconnect())

    async def _send_envelope_with_node_client(self, envelope: Envelope) -> None:
        """Send envelope with node client, reconnect and retry on fail."""
        if not self._node_client:  # pragma: nocover
            raise ValueError("Connection not connected to node!")

        self._ensure_valid_envelope_for_external_comms(envelope)
        try:
            await self._node_client.send_envelope(envelope)
        except Exception:  # pylint: disable=broad-except
            self.logger.exception(
                "Exception raised on message send. Try reconnect and send again."
            )
            await self._perform_connection_to_node()
            await self._node_client.send_envelope(envelope)

    async def connect(self) -> None:
        """Set up the connection."""
        if self.is_connected:  # pragma: nocover
            return

        with self._connect_context():
            # connect libp2p client

            await self._perform_connection_to_node()
            # start receiving msgs
            self._in_queue = asyncio.Queue()
            self._process_messages_task = asyncio.ensure_future(
                self._process_messages(), loop=self.loop
            )
            self._send_queue = asyncio.Queue()
            self._send_task = self.loop.create_task(self._send_loop())

    async def _perform_connection_to_node(self) -> None:
        """Connect to node with retries."""
        for attempt in range(self.connect_retries):
            if self.state not in [
                ConnectionStates.connecting,
                ConnectionStates.connected,
            ]:
                # do nothing if disconnected, or disconnecting
                return  # pragma: nocover
            try:
                self.logger.info(
                    "Connecting to libp2p node {}. Attempt {}".format(
                        str(self.node_uri), attempt + 1
                    )
                )
                self._node_client = NodeClient(self.node_uri, self.node_por)
                await self._setup_connection()

                self.logger.info(
                    "Successfully connected to libp2p node {}".format(
                        str(self.node_uri)
                    )
                )
                return
            except Exception as e:  # pylint: disable=broad-except
                if attempt == self.connect_retries - 1:
                    self.logger.error(
                        "Connection to  libp2p node {} failed: error: {}. It was the last attempt, exception will be raised".format(
                            str(self.node_uri), str(e)
                        )
                    )
                    self.state = ConnectionStates.disconnected
                    raise
                sleep_time = attempt * 2 + 1
                self.logger.error(
                    "Connection to  libp2p node {} failed: error: {}. Another attempt will be performed in {} seconds".format(
                        str(self.node_uri), str(e), sleep_time
                    )
                )
                await asyncio.sleep(sleep_time)

    async def _setup_connection(self) -> None:
        """Set up connection to node over tcp connection."""
        if not self._node_client:  # pragma: nocover
            raise ValueError("Connection was not connected!")

        await self._node_client.connect()

    async def disconnect(self) -> None:
        """Disconnect from the channel."""
        if self.is_disconnected:  # pragma: nocover
            return

        self.state = ConnectionStates.disconnecting
        self.logger.debug("disconnecting libp2p client connection...")

        if self._process_messages_task is not None:
            if not self._process_messages_task.done():
                self._process_messages_task.cancel()
            self._process_messages_task = None

        if self._send_task is not None:
            if not self._send_task.done():
                self._send_task.cancel()
            self._send_task = None

        try:
            self.logger.debug("disconnecting libp2p node client connection...")
            if self._node_client is not None:
                await self._node_client.close()
        except Exception:  # pragma: nocover  # pylint:disable=broad-except
            self.logger.exception("exception on node client close")
            raise
        finally:
            # set disconnected state anyway
            if self._in_queue is not None:
                self._in_queue.put_nowait(None)

            self.state = ConnectionStates.disconnected
            self.logger.debug("libp2p client connection disconnected.")

    async def receive(self, *args: Any, **kwargs: Any) -> Optional["Envelope"]:
        """
        Receive an envelope. Blocking.

        :param args: positional arguments
        :param kwargs: keyword arguments
        :return: the envelope received, or None.
        """
        try:
            if self._in_queue is None:
                raise ValueError("Input queue not initialized.")  # pragma: nocover
            envelope = await self._in_queue.get()
            if envelope is None:  # pragma: no cover
                self.logger.debug("Received None.")
                return None
            self.logger.debug("Received envelope: {}".format(envelope))
            return envelope
        except CancelledError:  # pragma: no cover
            self.logger.debug("Receive cancelled.")
            return None
        except Exception as e:  # pragma: no cover # pylint: disable=broad-except
            self.logger.exception(e)
            return None

    async def send(self, envelope: Envelope) -> None:
        """
        Send messages.

        :param envelope: the envelope
        """
        if not self._node_client or not self._send_queue:
            raise ValueError("Node is not connected!")  # pragma: nocover

        self._ensure_valid_envelope_for_external_comms(envelope)
        await self._send_queue.put(envelope)

    async def _read_envelope_from_node(self) -> Optional[Envelope]:
        """Read envelope from node, reconnec on error."""
        if not self._node_client:  # pragma: nocover
            raise ValueError("Connection not connected to node!")

        try:
            self.logger.debug("Waiting for messages...")
            envelope = await self._node_client.read_envelope()
            return envelope
        except ConnectionError as e:  # pragma: nocover
            self.logger.error(f"Connection error: {e}. Try to reconnect and read again")
        except IncompleteReadError as e:  # pragma: no cover
            self.logger.error(
                "Connection disconnected while reading from node ({}/{})".format(
                    len(e.partial), e.expected
                )
            )
        except Exception as e:  # pylint: disable=broad-except  # pragma: nocover
            self.logger.exception(f"On envelope read: {e}")

        try:
            self.logger.debug("Read envelope retry! Reconnect first!")
            await self._perform_connection_to_node()
            envelope = await self._node_client.read_envelope()
            return envelope  # pragma: no cover
        except Exception:  # pragma: no cover  # pylint: disable=broad-except
            self.logger.exception("Failed to read with reconnect!")
            return None

    async def _process_messages(self) -> None:
        """Receive data from node."""
        if not self._node_client:  # pragma: nocover
            raise ValueError("Connection not connected to node!")

        while True:
            envelope = await self._read_envelope_from_node()
            if self._in_queue is None:
                raise ValueError("Input queue not initialized.")  # pragma: nocover
            self._in_queue.put_nowait(envelope)
            if envelope is None:
                break  # pragma: no cover


class SSLValidator:
    """Interprocess communication channel client using tcp sockets with TLS."""

    def __init__(
        self,
        url: str,
        server_pub_key: str,
        logger: logging.Logger = _default_logger,
    ) -> None:
        """
        Check ssl certificate with server pub key.

        :param url: url to get signature
        :param server_pub_key: str, server public key to verify identity
        :param logger: the logger
        """
        self.server_pub_key = server_pub_key
        o = urlparse(url)
        self.url = url
        self.logger = logger
        self.host: str = cast(str, o.hostname)
        self.port: int = cast(int, o.port)

    async def check(self) -> ssl.SSLContext:
        """Check ssl/pubkey for mailbox service and return ssl context."""
        ssl_ctx, session_pub_key = await self.get_ssl_ctx_and_session_pub_key(
            self.host, self.port
        )
        signature = await self.get_signature(ssl_ctx)
        self._verify_session_key_signature(
            self.server_pub_key, signature, session_pub_key
        )
        return ssl_ctx

    async def get_signature(self, ssl_ctx: ssl.SSLContext) -> bytes:
        """
        Get signature for mailbox service (ssl pubkey signed with node private key).

        :param ssl_ctx: ssl context.

        :return: signature in bytes
        """
        async with aiohttp.ClientSession() as client:
            async with client.get(self.url, ssl=ssl_ctx) as resp:
                if resp.status != 200:  # pragma: nocover
                    raise ValueError("Bad server response")
                signature = await resp.read()
        return signature

    @staticmethod
    def _get_session_pub_key(writer: StreamWriter) -> bytes:  # pragma: nocover
        """Get session public key from tls stream writer."""
        cert_data = writer.get_extra_info("ssl_object").getpeercert(binary_form=True)

        cert = x509.Certificate.load(cert_data)
        session_pub_key = VerifyingKey.from_der(cert.public_key.dump()).to_string(
            "uncompressed"
        )
        return session_pub_key

    async def get_ssl_ctx_and_session_pub_key(
        self, host: str, port: int
    ) -> Tuple[ssl.SSLContext, bytes]:
        """Open a connection with TLS support."""
        cadata = await asyncio.get_event_loop().run_in_executor(
            None, lambda: ssl.get_server_certificate((host, port))
        )

        ssl_ctx = ssl.create_default_context(cadata=cadata)
        ssl_ctx.check_hostname = False
        ssl_ctx.verify_mode = ssl.CERT_REQUIRED
        _, writer = await asyncio.open_connection(
            self.host,
            self.port,
            ssl=ssl_ctx,
        )
        session_pub_key = self._get_session_pub_key(writer)
        writer.close()
        return ssl_ctx, session_pub_key

    @staticmethod
    def _verify_session_key_signature(
        server_pub_key: str, signature: bytes, session_pub_key: bytes
    ) -> None:
        """
        Validate signature of session public key.

        :param server_pub_key: node pub key/addr.
        :param signature: bytes, signature of session public key made with server private key
        :param session_pub_key: session public key to check signature for.
        """
        vk = VerifyingKey.from_string(bytes.fromhex(server_pub_key), SECP256k1)
        vk.verify(
            signature, session_pub_key, hashfunc=hashlib.sha256, sigdecode=sigdecode_der
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test package utils."""
import os
import re
from pathlib import Path
from typing import cast
from unittest import mock
from unittest.mock import patch

import click
import pytest

from aea import Version
from aea.cli.utils.context import Context
from aea.cli.utils.package_utils import (
    is_item_with_hash_present,
    list_available_packages,
    try_get_item_source_path,
    try_get_item_target_path,
    update_aea_version_range,
    validate_package_name,
    verify_private_keys_ctx,
)
from aea.configurations.base import AgentConfig
from aea.configurations.data_types import PublicId
from aea.configurations.manager import AgentConfigManager
from aea.helpers.ipfs.base import IPFSHashOnly
from aea.test_tools.test_cases import BaseAEATestCase


def test_verify_private_keys_ctx_fail() -> None:
    """Test `verify_private_keys_ctx`"""

    ctx = Context(".", "INFO", ".")
    with mock.patch.object(
        AgentConfigManager,
        "verify_private_keys",
        side_effect=ValueError,
    ):
        with pytest.raises(click.ClickException):
            verify_private_keys_ctx(ctx)

    with pytest.raises(
        AttributeError,
        match=re.escape("'Context' object has no attribute 'agent_config'"),
    ):
        _ = ctx.agent_config


def test_validate_package_name() -> None:
    """Test `validate_package_name`"""

    with pytest.raises(
        click.BadParameter, match=re.escape("bad-param is not a valid package name.")
    ):
        validate_package_name("bad-param")


def test_try_get_item_source_path() -> None:
    """Test `try_get_item_source_path`"""

    path = "."
    author_name = "author"
    item_type_plural = "skill"
    item_name = "some_skill"

    with mock.patch("os.path.exists", return_value=True):
        source_path = try_get_item_source_path(
            path, author_name, item_type_plural, item_name
        )
        assert source_path == os.path.join(
            path, author_name, item_type_plural, item_name
        )

    with pytest.raises(
        click.ClickException,
        match=re.escape(
            f'Item "{author_name}/{item_name}" not found in source folder "{source_path}".'
        ),
    ):
        source_path = try_get_item_source_path(
            path, author_name, item_type_plural, item_name
        )

    with mock.patch("os.path.exists", return_value=True):
        source_path = try_get_item_source_path(path, None, item_type_plural, item_name)
        assert source_path == os.path.join(path, item_type_plural, item_name)


def test_try_get_item_target_path() -> None:
    """Test `try_get_item_target_path`"""

    path = "."
    author_name = "author"
    item_type_plural = "skill"
    item_name = "some_skill"

    assert try_get_item_target_path(
        path, author_name, item_type_plural, item_name
    ) == os.path.join(path, author_name, item_type_plural, item_name)

    with mock.patch("os.path.exists", return_value=True):
        with pytest.raises(
            click.ClickException,
            match=re.escape('Item "some_skill" already exists in target folder'),
        ):
            try_get_item_target_path(path, author_name, item_type_plural, item_name)


class TestListAvailablePackages(BaseAEATestCase):
    """Test list_available_packages method."""

    use_packages_dir: bool = True

    def test_list_available_packages(
        self,
    ) -> None:
        """Test `list_available_packages`"""

        agent_name = "agent"
        self.create_agents(agent_name)

        package_list = list_available_packages(Path(self.t / agent_name))
        assert len(package_list) == 1

        (package_id, _), *_ = package_list
        assert package_id.name == "signing"

        with mock.patch.object(Path, "is_dir", return_value=False):
            package_list = list_available_packages(Path(self.t / agent_name))
            assert len(package_list) == 1

            (package_id, package_path), *_ = package_list
            assert package_id.name == "signing"
            assert "vendor" not in package_path.parts


def test_update_aea_version_range() -> None:
    """Test `update_aea_version_range`"""
    package_config = AgentConfig("name", "author", aea_version="0.1.0")
    versions = [
        "1.0.0",
        "2.0.0",
        "2.2.2",
        "1.30.0post0",
        "1.30.0.post1",
        "1.30.0dev1",
        "1.30.0.dev1",
        "0.0.1",
        "1000.0.10000",
        "1.1.0alpha0",
        "1.1.0.alpha1",
    ]
    for version_str in versions:
        version = Version(version_str)
        with patch(
            "aea.cli.utils.package_utils.get_current_aea_version", return_value=version
        ):
            update_aea_version_range(package_configuration=package_config)
            assert version in package_config.aea_version_specifiers


class TestItemPresentWithHash(BaseAEATestCase):
    """Test is_item_with_hash_present method."""

    use_packages_dir: bool = True

    def test_is_item_with_hash_present(
        self,
    ) -> None:
        """Test `is_item_with_hash_present`"""

        agent_name = "agent"
        self.create_agents(agent_name)

        agent_config = self.load_agent_config(agent_name)
        (signing_protocol,) = agent_config.protocols

        assert is_item_with_hash_present(
            str(self.packages_dir_path), agent_config, signing_protocol.hash
        )

        assert (
            is_item_with_hash_present(str(self.packages_dir_path), agent_config, "Qm")
            is None
        )

        dummy_config = mock.MagicMock()

        with mock.patch.object(
            IPFSHashOnly, "hash_directory", return_value="Qm"
        ), mock.patch(
            "aea.cli.utils.config.load_item_config", return_value=dummy_config
        ):
            assert (
                cast(
                    PublicId,
                    is_item_with_hash_present(
                        str(self.t / agent_name), agent_config, "Qm"
                    ),
                ).name
                == "signing"
            )
